# coding: utf-8
# /*##########################################################################
#
# Copyright (c) 2016-2017 European Synchrotron Radiation Facility
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
#
# ###########################################################################*/

__authors__ = ["H. Payno"]
__license__ = "MIT"
__date__ = "14/10/2021"


from silx.gui import qt
from tomwer.core.futureobject import FutureTomwerObject
from collections import OrderedDict
import functools


class FutureTomwerScanObserverWidget(qt.QWidget):
    """Widget used to observe a set of :class:`FutureTomwerScan`"""

    sigConversionPolicyChanged = qt.Signal()
    """Signal emit when the user change rule for converting finished scan"""

    sigConversionRequested = qt.Signal(object)
    """signal emit when a conversion is requested. Parameter is the future scan"""

    def __init__(self, parent=None):
        super().__init__(parent)
        self.setLayout(qt.QVBoxLayout())
        # filter widget
        self.filterWidget = FilterWidget(parent=self)
        self.layout().setContentsMargins(4, 4, 4, 4)
        self.layout().setSpacing(4)
        self.filterWidget.layout().setSpacing(4)
        self.layout().addWidget(self.filterWidget)
        # observation table
        self.observationTable = ObservationTable(self)
        self.layout().addWidget(self.observationTable)
        self.observationTable.setModel(
            _DatasetProcessModel(parent=self.observationTable, header=tuple())
        )
        self.observationTable.setSelectionBehavior(qt.QAbstractItemView.SelectRows)
        self.observationTable.resizeColumnsToContents()
        self.observationTable.setSortingEnabled(True)
        self.setSizePolicy(qt.QSizePolicy.Expanding, qt.QSizePolicy.Expanding)
        # converter option
        self._commandWidget = qt.QWidget(self)
        self.layout().addWidget(self._commandWidget)
        self._commandWidget.setLayout(qt.QGridLayout())
        self._commandWidget.setContentsMargins(0, 0, 0, 0)
        spacer = qt.QWidget(self._commandWidget)
        spacer.setSizePolicy(qt.QSizePolicy.Expanding, qt.QSizePolicy.Minimum)
        self._commandWidget.layout().addWidget(spacer, 0, 0, 1, 1)
        self._convertWhenFinished = qt.QCheckBox(
            "convert back to 'data' when finished", self
        )
        self._commandWidget.layout().addWidget(self._convertWhenFinished, 0, 1, 1, 2)
        self._convertSelectedPB = qt.QPushButton(self)
        self._convertSelectedPB.setText("convert selected")
        self._commandWidget.layout().addWidget(self._convertSelectedPB, 1, 2, 1, 1)
        self._cancelSelectedPB = qt.QPushButton(self)
        self._cancelSelectedPB.setText("cancel selected")
        style = qt.QApplication.style()
        stopIcon = style.standardIcon(qt.QStyle.SP_BrowserStop)
        self._cancelSelectedPB.setIcon(stopIcon)
        self._commandWidget.layout().addWidget(self._cancelSelectedPB, 1, 1, 1, 1)

        # set up
        self._convertWhenFinished.setChecked(True)
        h_header = self.observationTable.horizontalHeader()
        for column_index in (0, 2, 3, 4):
            h_header.setSectionResizeMode(column_index, qt.QHeaderView.Interactive)
        for column_index in (1,):
            h_header.setSectionResizeMode(column_index, qt.QHeaderView.Stretch)

        # connect signal / slot
        self.filterWidget.sigChanged.connect(self._filterUpdated)
        self._convertWhenFinished.toggled.connect(self._conversionPolicyChanged)
        self._convertSelectedPB.released.connect(self._convertSelected)
        self._cancelSelectedPB.released.connect(self._cancelSelected)

    def _convertSelected(self, *args, **kwargs):
        for cell in self.observationTable.selectionModel().selectedRows():
            future_tomo_obj = self.observationTable.model().futureTomoObjs.get(
                cell.row(), None
            )
            if future_tomo_obj is not None:
                self.sigConversionRequested.emit(future_tomo_obj)
        self.observationTable.model().layoutChanged.emit()

    def _cancelSelected(self, *args, **kwargs):
        for cell in self.observationTable.selectionModel().selectedRows():
            future_tomo_obj = self.observationTable.model().futureTomoObjs.get(
                cell.row(), None
            )
            if future_tomo_obj is not None:
                future_tomo_obj.cancel()
        self.observationTable.model().layoutChanged.emit()

    def addFutureTomoObj(self, future_tomo_obj: FutureTomwerObject):
        self.observationTable.model().addFutureTomoObj(future_tomo_obj)

    def removeFutureTomoObj(self, future_tomo_obj: FutureTomwerObject):
        self.observationTable.model().removeFutureTomoObj(future_tomo_obj)

    def _filterUpdated(self):
        self.observationTable.model().filtered_status = (
            self.filterWidget.getFilteredStatus()
        )

    def updateView(self):
        """
        Update scan view and each future scan status
        """
        self.observationTable.model().updateStatusAllFutureTomoObjs()
        self.observationTable.model().layoutChanged.emit()

    def convertWhenFinished(self):
        return self._convertWhenFinished.isChecked()

    def _conversionPolicyChanged(self, *args, **kwargs):
        self.sigConversionPolicyChanged.emit()

    def futureTomoObjs(self):
        return self.observationTable.model().futureTomoObjs

    def setConvertWhenFinished(self, convert):
        self._convertWhenFinished.setChecked(convert)


class _DatasetProcessModel(qt.QAbstractTableModel):
    sigStatusUpdated = qt.Signal(object, str)
    """Signal emit when the status of scan changes"""

    def __init__(self, parent, header, *args):
        qt.QAbstractTableModel.__init__(self, parent, *args)
        self.header = header
        self.futureTomoObjs = OrderedDict()
        # key is index, value is scan
        self._filteredFutureTomoObjs = OrderedDict()
        # key is index, value is scan but only for the scan we want to display
        self._tomoObjStatus = dict()
        # key is the future scan, value is the status of the scan
        self._filteredStatus = tuple()
        # list of status to filter
        """FutureTomwerScan with this status will be hide"""
        self._headers = {
            0: "start-time",
            1: "dataset",
            2: "status",
            3: "requester",
            4: "slurm-job",
            5: "info",
        }
        self._headersI = {v: k for k, v in self._headers.items()}
        assert len(self._headers) == len(self._headersI), "insure values are unique"

    def is_filtered(self, future_scan):
        return future_scan.status in self._filteredStatus

    def _computeUnfilteredFutureTomoObjs(self):
        i_row = 0
        self._filteredFutureTomoObjs = OrderedDict()

        for _, scan in self.futureTomoObjs.items():
            if not self.is_filtered(scan):
                self._filteredFutureTomoObjs[i_row] = scan
                i_row += 1
        return self._filteredFutureTomoObjs

    def endResetModel(self, *args, **kwargs):
        self._computeUnfilteredFutureTomoObjs()
        super().endResetModel(*args, **kwargs)

    def addFutureTomoObj(self, future_tomo_obj):
        if not isinstance(future_tomo_obj, FutureTomwerObject):
            raise TypeError(
                f"future_tomo_obj is expected to be an instance of {FutureTomwerObject} and not {type(future_tomo_obj)}"
            )
        self.futureTomoObjs[len(self.futureTomoObjs)] = future_tomo_obj
        self._tomoObjStatus[future_tomo_obj] = "pending"
        for future in future_tomo_obj.futures:
            callback = functools.partial(
                self._updateStatus,
                future_tomo_obj=future_tomo_obj,
            )

            # TODO: remove this since we are not using distributed anymore...
            class CallBack:
                # we cannot create a future directly because distributed enforce
                # the callback to have a function signature with only the future
                # as single parameter.
                def __init__(self, f_partial) -> None:
                    self.f_partial = f_partial

                def process(self, fn):
                    self.f_partial()

            future.add_done_callback(CallBack(callback).process)
        obj_index = len(self._filteredFutureTomoObjs)

        if not self.is_filtered(future_tomo_obj):
            self._filteredFutureTomoObjs[obj_index] = future_tomo_obj

        self.dataChanged.emit(
            self.createIndex(obj_index, 0),
            self.createIndex(obj_index, len(self._headers) - 1),
        )
        self.layoutChanged.emit()

    def updateStatusAllFutureTomoObjs(self):
        for future_tomo_obj in self._tomoObjStatus:
            self._tomoObjStatus[future_tomo_obj] = future_tomo_obj.status

    def _updateStatus(self, future_tomo_obj):
        if future_tomo_obj is None:
            return
        status = future_tomo_obj.status

        trigger_update = (
            future_tomo_obj not in self._tomoObjStatus
            or status != self._tomoObjStatus[future_tomo_obj]
            or status
            in (
                "finished",
                "completed",
            )  # hack: when several finish at the same time, ensure the conversion is done :()
        )
        if trigger_update:
            self._tomoObjStatus[future_tomo_obj] = status
            try:
                self.sigStatusUpdated.emit(future_tomo_obj, status)
            except Exception:
                pass
            else:
                obj_index = list(self._tomoObjStatus.keys()).index(future_tomo_obj)
                self.dataChanged.emit(
                    self.createIndex(obj_index, 0),
                    self.createIndex(obj_index, len(self._headers) - 1),
                )

    def removeFutureTomoObj(self, future_tomo_obj):
        for container in (self.futureTomoObjs, self._filteredFutureTomoObjs):
            futureTomoObjsI = {v: k for k, v in container.items()}
            idx = futureTomoObjsI.get(future_tomo_obj, None)
            if idx is not None:
                del container[idx]
        if future_tomo_obj in self._tomoObjStatus:
            del self._tomoObjStatus[future_tomo_obj]
        self.layoutChanged.emit()

    def getUnfilteredFutureTomoObjs(self):
        return self._filteredFutureTomoObjs

    @property
    def filtered_status(self):
        return self._filteredStatus

    @filtered_status.setter
    def filtered_status(self, filters):
        self._filteredStatus = filters
        self.layoutChanged.emit()

    def clear(self):
        self.futureTomoObjs = OrderedDict()
        self._filteredFutureTomoObjs = OrderedDict()
        self.layoutChanged.emit()

    def rowCount(self, parent=None):
        return len(self.getUnfilteredFutureTomoObjs())

    def columnCount(self, parent=None):
        return len(self._headers)

    def headerData(self, col, orientation, role):
        if orientation == qt.Qt.Horizontal and role == qt.Qt.DisplayRole:
            if col < len(self._headers):
                return self._headers[col]
            else:
                return None

    def data(self, index, role):
        if not index.isValid():
            return None

        if role not in (qt.Qt.DisplayRole, qt.Qt.ToolTipRole, qt.Qt.BackgroundRole):
            return None

        key = list(self._filteredFutureTomoObjs.keys())[index.row()]
        f_scan = self._filteredFutureTomoObjs[key]

        if role == qt.Qt.BackgroundRole:
            if index.column() in (self._headersI["status"],):
                # color are the same as the one defined by the process manager

                if f_scan.status == "error":
                    return qt.QColor("#f52718")
                elif f_scan.status == "pending":
                    return qt.QColor("#609ab3")
                elif f_scan.status == "cancelled":
                    return qt.QColor("#fcba03")
                elif f_scan.status == "running":
                    return qt.QColor("#839684")
                elif f_scan.status in ("finished", "completed"):
                    return qt.QColor("#068c0c")

        elif role == qt.Qt.DisplayRole:
            col_type = self._headers[index.column()]
            if col_type == "start-time":
                return f_scan.start_time.strftime("%H:%M:%S")
            elif col_type == "dataset":
                return str(f_scan.tomo_obj.get_identifier())
            elif col_type == "requester":
                from processview.core.manager import ProcessManager

                id_ = f_scan.process_requester_id
                process = ProcessManager().get_process(id_)
                if process is not None:
                    return process.name
                else:
                    return "unknown"
            elif col_type == "status":
                return f_scan.status
            elif col_type == "slurm-job":
                infos_job = ",".join(
                    [
                        str(future.job_id) if hasattr(future, "job_id") else ""
                        for future in f_scan.futures
                    ]
                )
                return infos_job


class ObservationTable(qt.QTableView):
    def __init__(self, parent):
        qt.QTableView.__init__(self, parent)

        # QMenu
        self.menu = qt.QMenu(self)


class FilterWidget(qt.QGroupBox):
    """
    Interface to define which future to be displayed
    """

    sigChanged = qt.Signal()
    """signal emit when filter change"""

    def __init__(self, parent=None):
        super().__init__("show job with status", parent)
        self.setLayout(qt.QHBoxLayout())
        # spacer
        spacer = qt.QWidget(self)
        spacer.setSizePolicy(qt.QSizePolicy.Expanding, qt.QSizePolicy.Minimum)
        self.layout().addWidget(spacer)

        # pending
        self._pending = qt.QCheckBox("pending", self)
        self.layout().addWidget(self._pending)
        # running
        self._running = qt.QCheckBox("running", self)
        self.layout().addWidget(self._running)
        # finished
        self._finished = qt.QCheckBox("finished", self)
        self.layout().addWidget(self._finished)
        # error
        self._error = qt.QCheckBox("error", self)
        self.layout().addWidget(self._error)
        # cancelled
        self._cancelled = qt.QCheckBox("cancelled", self)
        self.layout().addWidget(self._cancelled)

        # set up
        self._pending.setChecked(True)
        self._cancelled.setChecked(True)
        self._running.setChecked(True)
        self._finished.setChecked(True)
        self._error.setChecked(True)

        # connect signal / slot
        self._pending.toggled.connect(self._updated)
        self._cancelled.toggled.connect(self._updated)
        self._running.toggled.connect(self._updated)
        self._finished.toggled.connect(self._updated)
        self._error.toggled.connect(self._updated)

    def _updated(self):
        self.sigChanged.emit()

    def getFilteredStatus(self):
        filters = []
        cbs_filter = {
            "pending": self._pending,
            "cancelled": self._cancelled,
            "running": self._running,
            "finished": self._finished,
            "error": self._error,
        }
        for status, status_cb in cbs_filter.items():
            if not status_cb.isChecked():
                filters.append(status)
        return tuple(filters)



# coding: utf-8
# /*##########################################################################
#
# Copyright (c) 2016-2017 European Synchrotron Radiation Facility
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
#
# ###########################################################################*/

__authors__ = ["H. Payno"]
__license__ = "MIT"
__date__ = "01/06/2018"


from collections import namedtuple

_getterSetter = namedtuple("_getterSetter", ["getter", "setter"])


class _TofuOptionLoader(object):
    """Simple class used to register the options managed by a TofuXWidget"""

    def __init__(self, options, childs=None):
        self._options = options
        self._childs = childs or []
        # expose API
        self._addChild = self._childs.append

    def _hasOption(self, option):
        return option in self._options

    def _setOption(self, option, value):
        assert option in self._options
        if value == "":
            self._options[option].setter()
        else:
            self._options[option].setter(value)

    def getParameters(self):
        """Add the value of the detain options + some extra parameters to be
        passed as 'free string'

        :return:
        :rtype: tuple (dict, list)
        """
        _ddict, extraParams = self._getNodeParameters()
        for child in self._childs:
            _childDict = child.getParameters()
            _ddict.update(_childDict)
        return _ddict

    def _getNodeParameters(self):
        """Add the value of the detain options + some extra parameters to be
        passed as 'free string'

        :return:
        :rtype: tuple (dict, list)
        """
        _ddict = {}
        for option in self._options:
            _ddict[option] = self._options[option].getter()
        return _ddict, []

    def setParameters(self, _ddict):
        """

        :param _ddict:
        :return: initial dictionary less the parameters managed by this option
                 loader
         :rtype dict:
        """
        _lddict = _ddict.copy()
        for child in self._childs:
            _lddict = child.setParameters(_lddict)
        _lddict = self._setNodeParameters(_lddict)
        return _lddict

    def _setNodeParameters(self, _ddict):
        """

        :param _ddict:
        :return: initial dictionary less the parameters managed by this option
                 loader
         :rtype dict:
        """
        updatedDict = _ddict.copy()
        for param, value in _ddict.items():
            if param in self._options:
                self._options[param].setter(value)
                del updatedDict[param]
        return updatedDict



# coding: utf-8
# /*##########################################################################
# Copyright (C) 2016 European Synchrotron Radiation Facility
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
#
#############################################################################*/

__authors__ = ["H. Payno"]
__license__ = "MIT"
__date__ = "22/01/2017"


import os
from tomwer.tests.utils import skip_gui_test
from silx.gui import qt
from silx.gui.utils.testutils import TestCaseQt
from tomwer.gui.control.volumeselectorwidget import VolumeSelectorWidget
from tomwer.core.volume import EDFVolume, HDF5Volume
import numpy
import shutil
import tempfile
import pytest


@pytest.mark.skipif(skip_gui_test(), reason="skip gui test")
class TestVolumeSelector(TestCaseQt):
    """
    Simple test for the VolumeSelector
    """

    def setUp(self):
        super().setUp()
        self.folder = tempfile.mkdtemp()

        self.volume_1 = HDF5Volume(
            file_path=os.path.join(self.folder, "my_volume.hdf5"),
            data_path="entry0000",
            data=numpy.linspace(start=1, stop=100, num=300).reshape((3, 10, 10)),
        )
        self.volume_1.save()

        self.volume_2 = HDF5Volume(
            file_path=os.path.join(self.folder, "my_volume.hdf5"),
            data_path="entry0001",
            data=numpy.linspace(start=1, stop=100, num=400).reshape((4, 10, 10)),
        )
        self.volume_2.save()

        self.volume_3 = EDFVolume(
            folder=os.path.join(self.folder, "vol"),
            data=numpy.linspace(start=1, stop=100, num=500).reshape((5, 10, 10)),
        )
        self.volume_3.save()

        self.widget = VolumeSelectorWidget(parent=None)

    def tearDown(self):
        self.widget.setAttribute(qt.Qt.WA_DeleteOnClose)
        self.widget.close()
        self.widget = None
        shutil.rmtree(self.folder)
        super().tearDown()

    def test(self):
        for volume in (self.volume_1, self.volume_2, self.volume_3):
            self.widget.add(volume)

        # self.widget.show()
        # self.qapp.exec_()
        self.assertEqual(self.widget.dataList.n_data(), 3)
        self.widget.remove(self.volume_2)
        self.assertEqual(self.widget.dataList.n_data(), 2)
        self.widget.remove(self.volume_3)
        self.assertEqual(self.widget.dataList.n_data(), 1)
        assert self.widget.dataList.rowCount() == 1



# coding: utf-8
# /*##########################################################################
#
# Copyright (c) 2016-2017 European Synchrotron Radiation Facility
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
#
# ###########################################################################*/

__authors__ = ["H. Payno"]
__license__ = "MIT"
__date__ = "27/02/2021"


import unittest
from tomwer.core.process.reconstruction.scores.scores import compute_score_contrast_std
from tomwer.core.process.reconstruction.scores.scores import _METHOD_TO_FCT
from tomwer.core.process.reconstruction.saaxis.saaxis import SAAxisProcess
from tomwer.core.process.reconstruction.saaxis.params import SAAxisParams
from tomwer.core.utils.scanutils import MockHDF5
import os
import tempfile
import shutil
import numpy


class TestScoreFunctions(unittest.TestCase):
    """Test all the score functions"""

    def test_img_contrast_std_score(self):
        """simple test of the API to call compute_score_contrast_std"""
        data = numpy.random.random(100 * 100).reshape(100, 100)
        compute_score_contrast_std(data)

    def test_method_to_function(self):
        """Test the dictionary used to for linking the score method to the
        callback function"""
        data = numpy.random.random(100 * 100).reshape(100, 100)
        for method_name, fct in _METHOD_TO_FCT.items():
            with self.subTest(method_name=method_name, fct=fct):
                res = fct(data)
                self.assertFalse(res is None)
                self.assertTrue(isinstance(res, float))


class TestSAAxisProcess(unittest.TestCase):
    """Test the SAAxisProcess class"""

    def setUp(self) -> None:
        self.tempdir = tempfile.mkdtemp()
        dim = 10
        mock = MockHDF5(
            scan_path=self.tempdir, n_proj=10, n_ini_proj=10, scan_range=180, dim=dim
        )
        mock.add_alignment_radio(index=10, angle=90)
        mock.add_alignment_radio(index=10, angle=0)
        self.scan = mock.scan

        self._default_saaxis_params = SAAxisParams()
        self._default_saaxis_params.output_dir = os.path.join(
            self.tempdir, "output_dir"
        )
        self._default_saaxis_params.slice_indexes = {"slice": 4}
        self._default_saaxis_params.nabu_config = {}
        self._default_saaxis_params.dry_run = True
        self._default_saaxis_params.file_format = "hdf5"

    def tearDown(self) -> None:
        shutil.rmtree(self.tempdir)

    def test(self):
        process = SAAxisProcess(inputs={"data": self.scan})
        with self.assertRaises(Exception):
            process.run()
        process.set_configuration(self._default_saaxis_params)
        with self.assertRaises(ValueError):
            process.run()
        self._default_saaxis_params.estimated_cor = 11
        self._default_saaxis_params.research_width = 2
        process.set_configuration(self._default_saaxis_params)
        process.run()
        process = SAAxisProcess(
            inputs={"data": self.scan, "sa_axis_params": self._default_saaxis_params}
        )
        process.run()



# coding: utf-8
###########################################################################
# Copyright (C) 2016 European Synchrotron Radiation Facility
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
#
#############################################################################

"""
This module is dedicated to cast of volume from one file format to the other.
"""

__authors__ = ["H.Payno"]
__license__ = "MIT"
__date__ = "15/12/2021"


import os
from processview.core.superviseprocess import SuperviseProcess
from tomwer.core.cluster.cluster import SlurmClusterConfiguration
from tomwer.core.process.reconstruction.nabu import settings
from tomwer.core.process.task import Task
from tomwer.core.process.reconstruction.nabu.nabucommon import (
    NabuOutputFileFormat,
)
from processview.core.manager import ProcessManager, DatasetState
import numpy
import logging
from tomoscan.volumebase import VolumeBase
from nabu.io.cast_volume import (
    cast_volume as _nabu_cast_volume,
    RESCALE_MIN_PERCENTILE,
    RESCALE_MAX_PERCENTILE,
    get_default_output_volume,
)
from tomwer.core.futureobject import FutureTomwerObject
from tomwer.core.utils.slurm import get_slurm_script_name
from tomwer.core.volume.volumefactory import VolumeFactory
from tomwer.core.volume.volumebase import TomwerVolumeBase
from tomwer.core.utils.volumeutils import volume_identifier_to_volume
from sluurp.job import SBatchScriptJob
from sluurp.executor import submit as submit_to_slurm_cluster

_logger = logging.getLogger(__name__)


DEFAULT_OUTPUT_DIR = "{volume_data_parent_folder}/cast_volume"


class CastVolumeTask(
    Task,
    SuperviseProcess,
    input_names=(
        "volume",
        "configuration",
    ),
    optional_input_names=("scan", "output_volume"),
    output_names=(
        "volume",
        "future_tomo_obj",
    ),
):
    def __init__(
        self,
        process_id=None,
        varinfo=None,
        inputs=None,
        node_id=None,
        node_attrs=None,
        execinfo=None,
    ):
        SuperviseProcess.__init__(self, process_id=process_id)
        Task.__init__(
            self,
            varinfo=varinfo,
            inputs=inputs,
            node_id=node_id,
            node_attrs=node_attrs,
            execinfo=execinfo,
        )
        self._cluster_config = None

    def run(self):
        input_volume = volume_identifier_to_volume(self.inputs.volume)
        if not isinstance(input_volume, VolumeBase):
            raise TypeError(
                f"input_volume is a {type(input_volume)} when {VolumeBase} expected"
            )
        configuration = self.inputs.configuration
        self._cluster_config = configuration.pop("cluster_config", None)
        if isinstance(self._cluster_config, SlurmClusterConfiguration):
            self._cluster_config = self._cluster_config.to_dict()
        if self.inputs.scan:
            scan = self.inputs.scan
        else:
            scan = None
        if self.inputs.output_volume:
            output_volume = self.inputs.output_volume
            if not isinstance(output_volume, TomwerVolumeBase):
                output_volume = VolumeFactory.create_tomo_object_from_identifier(
                    output_volume
                )
        else:
            output_data_type = configuration.get(
                "output_type", None
            )  # expects values like hdf5, edf ...
            output_dir = configuration.get("output_dir", None)
            if output_data_type is None or output_dir is None:
                mess = "'output_volume' or ('output_type' and 'output_dir' from configuraiton) should be provided."
                _logger.processFailed(f"volume cast of {str(input_volume)} failed")
                ProcessManager().notify_dataset_state(
                    dataset=input_volume,
                    process=self,
                    state=DatasetState.FAILED,
                )
                raise ValueError(mess)

            output_dir = TomwerVolumeBase.format_output_location(
                location=output_dir,
                volume=input_volume,
            )

            output_volume = get_default_output_volume(
                input_volume=input_volume,
                output_type=NabuOutputFileFormat.from_value(output_data_type).value,
                output_dir=output_dir,
            )
            # convert from tomoscan volume to tomwer volume
            output_volume = VolumeFactory.create_tomo_object_from_identifier(
                output_volume.get_identifier().to_str()
            )

        # build output volume object
        overwrite = configuration.get("overwrite", False)
        output_volume.overwrite = overwrite

        output_data_type = numpy.dtype(
            configuration.get("output_data_type")
        )  # expects values like numpy uint8, unint16...

        rescale_min_percentile = configuration.get(
            "rescale_min_percentile", RESCALE_MIN_PERCENTILE
        )
        rescale_max_percentile = configuration.get(
            "rescale_max_percentile", RESCALE_MAX_PERCENTILE
        )
        data_min = configuration.get("data_min", None)
        data_max = configuration.get("data_max", None)

        # run volume cast locally
        if self._cluster_config is None:
            try:
                _nabu_cast_volume(
                    input_volume=input_volume,
                    output_volume=output_volume,
                    output_data_type=output_data_type,
                    data_min=data_min,
                    data_max=data_max,
                    scan=scan,
                    rescale_min_percentile=rescale_min_percentile,
                    rescale_max_percentile=rescale_max_percentile,
                    save=True,
                    store=False,
                )
            except Exception as e:
                mess = f"volume cast of {str(input_volume)} failed. Reason is {str(e)}"
                _logger.processFailed(mess)
                state = DatasetState.FAILED
            else:
                mess = f"volume cast of {str(input_volume)} succeed"
                _logger.processSucceed(mess)
                state = DatasetState.SUCCEED

            ProcessManager().notify_dataset_state(
                dataset=input_volume,
                process=self,
                state=state,
                details=mess,
            )
            self.outputs.future_tomo_obj = None
            self.outputs.volume = output_volume
            if scan is not None:
                scan.cast_volume = output_volume.get_identifier()
            else:
                input_volume.cast_volume = output_volume.get_identifier()
        # run volume cast remotly
        else:

            def get_command():
                command = f"python3 -m {settings.NABU_CAST_APP_PATH} '{input_volume.get_identifier().to_str()}'"
                command += (
                    f" --output_volume='{output_volume.get_identifier().to_str()}'"
                )
                if overwrite:
                    command += " --overwrite"
                command += f" --output_type={str(output_data_type)}"
                if data_min is not None:
                    command += f" --data_min={data_min}"
                if data_max is not None:
                    command += f" --data_max={data_max}"
                if rescale_min_percentile is not None:
                    command += f" --rescale_min_percentile={rescale_min_percentile}"
                if rescale_max_percentile is not None:
                    command += f" --rescale_max_percentile={rescale_max_percentile}"

                return command

            script_name = get_slurm_script_name(prefix="nabu_cast")
            if scan is not None:
                working_directory = scan.working_directory
                script_path = os.path.join(scan.path, "slurm_scripts", script_name)
            else:
                working_directory = os.path.dirname(input_volume.data_url.file_path())
                script_path = os.path.join(
                    working_directory, "slurm_scripts", script_name
                )
            # for now force job name
            self._cluster_config[
                "job_name"
            ] = f"tomwer-cast-volume {input_volume.get_identifier().to_str()} to {output_volume.get_identifier().to_str()}"
            job = SBatchScriptJob(
                slurm_config=self._cluster_config,
                script=(get_command(),),
                script_path=script_path,
                clean_script=False,
                working_directory=working_directory,
            )
            future_slurm_job = submit_to_slurm_cluster(job)
            if scan is not None:
                future_obj = scan
            else:
                future_obj = output_volume
            self.outputs.future_tomo_obj = FutureTomwerObject(
                tomo_obj=future_obj,
                futures=(future_slurm_job,),
                process_requester_id=self.process_id,
            )
            self.outputs.volume = output_volume



# coding: utf-8
# /*##########################################################################
#
# Copyright (c) 2016-2017 European Synchrotron Radiation Facility
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
#
# ###########################################################################*/

__authors__ = ["H. Payno"]
__license__ = "MIT"
__date__ = "21/07/2020"

from silx.gui import qt
from orangewidget import gui
from orangewidget.widget import Input, Output

import tomwer.core.process.control.scanvalidator
from tomwer.gui.control.datavalidator import DataValidator
from tomwer.core.scan.scanbase import TomwerScanBase
from orangecontrib.tomwer.orange.managedprocess import SuperviseOW
from processview.core.manager import ProcessManager, DatasetState
from tomwer.core.scan.scanbase import _TomwerBaseDock
from orangewidget import settings
from tomwer.utils import docstring
import logging

_logger = logging.getLogger(__name__)


class DataValidatorOW(SuperviseOW):
    """a data viewer able to:

    - display slices (latest reconstructed if any)
    - display radios with or without normalization

    :param parent: the parent widget
    """

    name = "scan validator"
    id = "orange.widgets.tomwer.datavalidator"
    description = """Widget displaying results of a reconstruction and asking to
        the user if he want to validate or not the reconstruction. User can also ask
        for some modification on the reconstruction parameters"""
    icon = "icons/validator.png"
    priority = 23
    keywords = ["tomography", "file", "tomwer", "acquisition", "validation"]

    want_main_area = True
    resizing_enabled = True
    compress_signal = False

    ewokstaskclass = tomwer.core.process.control.scanvalidator._ScanValidatorPlaceHolder

    _viewer_config = settings.Setting(dict())

    _warnValManualShow = False
    """
    used to know if the message to inform user about `validate manually` has
    already been displayed.
    This informative message will be show under the following conditions:

        * the scanValidator contains at least `_NB_SCAN_BF_WARN`
        * this dialog have never been showed in the current session.
    """

    _NB_SCAN_BF_WARN = 10
    """
    Limit of stored scans before displaying the informative message about
    `validate manually` checkbox
    """

    class Inputs:
        data = Input(name="data", type=TomwerScanBase)

    class Outputs:
        data = Output(name="data", type=TomwerScanBase)
        recons_params_changed = Output(
            name="change recons params",
            type=_TomwerBaseDock,
        )

    def __init__(self, parent=None):
        SuperviseOW.__init__(self, parent)
        self._layout = gui.vBox(self.mainArea, self.name).layout()
        self._widget = DataValidator(parent=self)
        self._layout.addWidget(self._widget)
        self._setSettings(settings=self._viewer_config)

        # connect signal / slots
        self._widget.sigChangeReconsParams.connect(self._changeReconsParamsEmited)
        self._widget.sigScanReady.connect(self._scanReadyEmitted)
        self._widget._centralWidget.sigConfigChanged.connect(self._updateSettings)

    def close(self):
        if self._widget is not None:
            self._widget.setAttribute(qt.Qt.WA_DeleteOnClose)
            self._widget.close()
            self._widget = None
        super().close()

    @Inputs.data
    def addScan(self, scan):
        if scan is None:
            return
        assert isinstance(scan, TomwerScanBase)
        self._widget.addScan(scan)
        ProcessManager().notify_dataset_state(
            dataset=scan, process=self, state=DatasetState.WAIT_USER_VALIDATION
        )
        # in the case the memory is full, the scan can have been already
        # validated and so not accessible
        if (
            self._warnValManualShow is False
            and len(self._widget._scans) >= self._NB_SCAN_BF_WARN
        ):
            mess = (
                "Please note that the scanValidator is actually storing %s "
                "scan(s). \n"
                "Scan need to be validated manually in order to continue "
                "the workflow processing. \n"
                "you can either validate scan manually or uncheck the "
                "`validate manually` check box." % self._NB_SCAN_BF_WARN
            )

            mess = qt.QMessageBox(self, qt.QMessageBox.Information, mess)
            mess.setModal(False)
            mess.show()
            self._warnValManualShow = True

        if self.isValidationManual():
            self.show()
            self.activateWindow()
            self.raise_()

    @docstring
    def reprocess(self, dataset):
        self.addScan(dataset)

    def isValidationManual(self):
        return self._widget.isValidationManual()

    def _changeReconsParamsEmited(self, scan):
        self.Outputs.recons_params_changed.send(scan)

    def _scanReadyEmitted(self, scan):
        ProcessManager().notify_dataset_state(
            dataset=scan, process=self, state=DatasetState.SUCCEED
        )
        self.Outputs.data.send(scan)

    def _updateSettings(self):
        viewer = self._widget._centralWidget
        self._viewer_config["mode"] = viewer.getDisplayMode()  # pylint: disable=E1137
        self._viewer_config[  # pylint: disable=E1137
            "slice_opt"
        ] = viewer.getSliceOption()
        self._viewer_config[  # pylint: disable=E1137
            "radio_opt"
        ] = viewer.getRadioOption()

    def _setSettings(self, settings):
        viewer = self._widget._centralWidget
        old_state = viewer.blockSignals(True)
        if "mode" in settings:
            viewer.setDisplayMode(settings["mode"])
        if "slice_opt" in settings:
            viewer.setSliceOption(settings["slice_opt"])
        if "radio_opt" in settings:
            viewer.setRadioOption(settings["radio_opt"])
        viewer.blockSignals(old_state)

    def getNScanToValidate(self):
        return len(self._widget._scans)

    def _validateScan(self, scan):
        self._widget._validateScan(scan=scan)

    def setAutomaticValidation(self, auto):
        self._widget.setAutomaticValidation(auto)



# coding: utf-8
# /*##########################################################################
#
# Copyright (c) 2016-2017 European Synchrotron Radiation Facility
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
#
# ###########################################################################*/

__authors__ = ["H. Payno"]
__license__ = "MIT"
__date__ = "31/03/2021"


from silx.gui import qt
from silx.io.url import DataUrl
from silx.io.utils import h5py_read_dataset
from tomwer.core.utils.char import BETA_CHAR, DELTA_CHAR
from tomoscan.io import HDF5File
import os


class ReconstructionParameters(qt.QWidget):
    def __init__(self, parent=None):
        qt.QWidget.__init__(self, parent)
        self.setLayout(qt.QFormLayout())
        # method
        self._methodQLE = qt.QLineEdit("", self)
        self.layout().addRow("method", self._methodQLE)
        # paganin
        self._paganinCB = qt.QCheckBox("", self)
        self._paganinCB.setEnabled(False)
        self.layout().addRow("Paganin", self._paganinCB)
        # delta beta
        self._deltaBetaQLE = qt.QLineEdit("", self)
        self._deltaBetaQLE.setReadOnly(True)
        self.layout().addRow(" / ".join((DELTA_CHAR, BETA_CHAR)), self._deltaBetaQLE)
        # distance
        self._distanceQLE = qt.QLineEdit("", self)
        self._distanceQLE.setReadOnly(True)
        self.layout().addRow("distance (cm)", self._distanceQLE)
        # pixel size
        self._pixelSizeQLE = qt.QLineEdit("", self)
        self._pixelSizeQLE.setReadOnly(True)
        self.layout().addRow("pixel size (cm)", self._pixelSizeQLE)
        # cor
        self._corQLE = qt.QLineEdit("", self)
        self._corQLE.setReadOnly(True)
        self.layout().addRow("cor (absolute)", self._corQLE)
        # padding type
        self._paddingTypeQLE = qt.QLineEdit("", self)
        self._paddingTypeQLE.setReadOnly(True)
        self.layout().addRow("padding type", self._paddingTypeQLE)
        # half tomo
        self._halfTomoCB = qt.QCheckBox("", self)
        self._halfTomoCB.setEnabled(False)
        self.layout().addRow("half tomo", self._halfTomoCB)
        # fbp filter type
        self._fbpFilterQLE = qt.QLineEdit("", self)
        self._fbpFilterQLE.setReadOnly(True)
        self.layout().addRow("fbp filter", self._fbpFilterQLE)
        # log min clip
        self._minLogClipQLE = qt.QLineEdit("", self)
        self._minLogClipQLE.setReadOnly(True)
        self.layout().addRow("log min clip", self._minLogClipQLE)
        # log max clip
        self._maxLogClipQLE = qt.QLineEdit("", self)
        self._maxLogClipQLE.setReadOnly(True)
        self.layout().addRow("log max clip", self._maxLogClipQLE)
        # sino normalization & normalization file
        self._sinonormalizationQLE = qt.QLabel("", self)
        self.layout().addRow("sino normalization", self._sinonormalizationQLE)
        self._sinonormalizationFileQLE = qt.QLabel("", self)
        self.layout().addRow("sino normalization file", self._sinonormalizationFileQLE)
        # software version
        self._softwareVersionQLE = qt.QLabel("", self)
        self.layout().addRow("software version", self._softwareVersionQLE)

    def setUrl(self, url):
        if isinstance(url, DataUrl):
            pass
        elif isinstance(url, str):
            url = DataUrl(path=url)
        else:
            raise TypeError(
                "url should be a DataUrl or a str representing"
                "an url and not {}".format(type(url))
            )
        for func in (
            self._setMethod,
            self._setPaganin,
            self._setDeltaBeta,
            self._setDistance,
            self._setPixelSize,
            self._setCor,
            self._setPaddingType,
            self._setHalfTomo,
            self._setFBPFilter,
            self._setMinLogClip,
            self._setMaxLogClip,
            self._setSinoNormalization,
            self._setSoftwareVersion,
        ):
            func(url)

    @staticmethod
    def _decode_nabu_str(input):
        if hasattr(input, "decode"):
            input = input.decode("UTF-8")
        return input

    def _setMethod(self, url):
        def get_method_value():
            _NABU_METHOD_URL = "../../configuration/nabu_config/reconstruction/method"
            with HDF5File(url.file_path(), "r") as h5s:
                data_path = url.data_path()
                method_path = os.path.normpath("/".join((data_path, _NABU_METHOD_URL)))
                if method_path in h5s:
                    return h5py_read_dataset(h5s[method_path])
                else:
                    return None

        value = self._decode_nabu_str(get_method_value())
        self._methodQLE.setText(value if value is not None else "")

    def _setPaganin(self, url):
        def is_phase_defined():
            _NABU_PHASE_URL = "../../configuration/nabu_config/phase"
            with HDF5File(url.file_path(), "r") as h5s:
                data_path = url.data_path()
                phase_path = os.path.normpath("/".join((data_path, _NABU_PHASE_URL)))
                if phase_path in h5s and "method" in h5s[phase_path]:
                    method_dataset = h5py_read_dataset(h5s[phase_path]["method"])
                    return method_dataset == "paganin"
                else:
                    return False

        self._paganinCB.setChecked(is_phase_defined())

    def _setDeltaBeta(self, url):
        def get_delta_beta_value():
            _NABU_DELTA_BETA_URL = "../../configuration/nabu_config/phase/delta_beta"
            with HDF5File(url.file_path(), "r") as h5s:
                data_path = url.data_path()
                delta_beta_path = os.path.normpath(
                    "/".join((data_path, _NABU_DELTA_BETA_URL))
                )
                if delta_beta_path in h5s:
                    return h5py_read_dataset(h5s[delta_beta_path])
                else:
                    return None

        value = self._decode_nabu_str(get_delta_beta_value())
        self._deltaBetaQLE.setText(str(value) if value is not None else "")

    def _setDistance(self, url):
        def get_distance_value():
            _NABU_DISTANCE_URL = (
                "../../configuration/processing_options/phase/distance_cm"
            )
            with HDF5File(url.file_path(), "r") as h5s:
                data_path = url.data_path()
                distance_path = os.path.normpath(
                    "/".join((data_path, _NABU_DISTANCE_URL))
                )
                if distance_path in h5s:
                    return h5py_read_dataset(h5s[distance_path])
                else:
                    return None

        value = self._decode_nabu_str(get_distance_value())
        if value is not None:
            value = "{:.2}".format(float(value))
        self._distanceQLE.setText(value if value is not None else "")

    def _setPixelSize(self, url):
        def get_pixel_size_value():
            _NABU_PIXEL_SIZE_URL = (
                "../../configuration/processing_options/reconstruction/pixel_size_cm"
            )
            with HDF5File(url.file_path(), "r") as h5s:
                data_path = url.data_path()
                pixel_size_path = os.path.normpath(
                    "/".join((data_path, _NABU_PIXEL_SIZE_URL))
                )
                if pixel_size_path in h5s:
                    return h5py_read_dataset(h5s[pixel_size_path])
                else:
                    return None

        value = self._decode_nabu_str(get_pixel_size_value())
        if value is not None:
            value = "{:.8}".format(float(value))
        self._pixelSizeQLE.setText(value if value is not None else "")

    def _setCor(self, url):
        def get_cor_value():
            _NABU_COR_URL = "../../configuration/processing_options/reconstruction/rotation_axis_position"
            with HDF5File(url.file_path(), "r") as h5s:
                data_path = url.data_path()
                pixel_size_path = os.path.normpath("/".join((data_path, _NABU_COR_URL)))
                if pixel_size_path in h5s:
                    return h5py_read_dataset(h5s[pixel_size_path])
                else:
                    return None

        value = self._decode_nabu_str(get_cor_value())
        if value not in (None, "None", "none"):
            value = "{:.4f}".format(float(value))
        self._corQLE.setText(value if value is not None else "")

    def _setPaddingType(self, url):
        def get_padding_type_value():
            _NABU_PADDING_URL = (
                "../../configuration/processing_options/reconstruction/padding_type"
            )
            with HDF5File(url.file_path(), "r") as h5s:
                data_path = url.data_path()
                padding_path = os.path.normpath(
                    "/".join((data_path, _NABU_PADDING_URL))
                )
                if padding_path in h5s:
                    return h5py_read_dataset(h5s[padding_path])
                else:
                    return None

        value = self._decode_nabu_str(get_padding_type_value())
        self._paddingTypeQLE.setText(str(value) if value is not None else "")

    def _setHalfTomo(self, url):
        def get_half_tomo_value():
            _NABU_HALF_TOMO_URL = (
                "../../configuration/processing_options/reconstruction/enable_halftomo"
            )
            with HDF5File(url.file_path(), "r") as h5s:
                data_path = url.data_path()
                half_tomo_path = os.path.normpath(
                    "/".join((data_path, _NABU_HALF_TOMO_URL))
                )
                if half_tomo_path in h5s:
                    return h5py_read_dataset(h5s[half_tomo_path])
                else:
                    return False

        value = get_half_tomo_value()
        self._halfTomoCB.setChecked(value)

    def _setFBPFilter(self, url):
        def get_fbp_filter_value():
            _NABU_FBP_FILTER_URL = (
                "../../configuration/processing_options/reconstruction/fbp_filter_type"
            )
            with HDF5File(url.file_path(), "r") as h5s:
                data_path = url.data_path()
                fbp_filter_path = os.path.normpath(
                    "/".join((data_path, _NABU_FBP_FILTER_URL))
                )
                if fbp_filter_path in h5s:
                    return h5py_read_dataset(h5s[fbp_filter_path])
                else:
                    return None

        value = self._decode_nabu_str(get_fbp_filter_value())
        self._fbpFilterQLE.setText(str(value) if value is not None else "")

    def _setMinLogClip(self, url):
        def get_log_min_clip_value():
            _NABU_LOG_MIN_CLIP_URL = (
                "../../configuration/processing_options/take_log/log_min_clip"
            )
            with HDF5File(url.file_path(), "r") as h5s:
                data_path = url.data_path()
                log_min_path = os.path.normpath(
                    "/".join((data_path, _NABU_LOG_MIN_CLIP_URL))
                )
                if log_min_path in h5s:
                    return h5py_read_dataset(h5s[log_min_path])
                else:
                    return None

        value = self._decode_nabu_str(get_log_min_clip_value())
        self._minLogClipQLE.setText(str(value) if value is not None else "")

    def _setMaxLogClip(self, url):
        def get_log_max_clip_value():
            _NABU_LOG_MAX_CLIP_URL = (
                "../../configuration/processing_options/take_log/log_max_clip"
            )
            with HDF5File(url.file_path(), "r") as h5s:
                data_path = url.data_path()
                log_max_path = os.path.normpath(
                    "/".join((data_path, _NABU_LOG_MAX_CLIP_URL))
                )
                if log_max_path in h5s:
                    return h5py_read_dataset(h5s[log_max_path])
                else:
                    return None

        value = self._decode_nabu_str(get_log_max_clip_value())
        self._maxLogClipQLE.setText(str(value) if value is not None else "")

    def _setSinoNormalization(self, url):
        def get_normalization_method():
            _NABU_NORM_URL = (
                "../../configuration/processing_options/sino_normalization/method"
            )
            with HDF5File(url.file_path(), "r") as h5s:
                data_path = url.data_path()
                log_max_path = os.path.normpath("/".join((data_path, _NABU_NORM_URL)))
                if log_max_path in h5s:
                    return h5py_read_dataset(h5s[log_max_path])
                else:
                    return None

        def get_normalization_file():
            _NABU_NORM_FILE_URL = (
                "../../configuration/nabu_config/preproc/sino_normalization_file"
            )
            with HDF5File(url.file_path(), "r") as h5s:
                data_path = url.data_path()
                log_max_path = os.path.normpath(
                    "/".join((data_path, _NABU_NORM_FILE_URL))
                )
                if log_max_path in h5s:
                    return h5py_read_dataset(h5s[log_max_path])
                else:
                    return None

        method = get_normalization_method()
        normalization_file = get_normalization_file()
        self._sinonormalizationQLE.setText(self._decode_nabu_str(method))
        self._sinonormalizationFileQLE.setText(
            self._decode_nabu_str(normalization_file)
        )
        self._sinonormalizationFileQLE.setToolTip(normalization_file)

    def _setSoftwareVersion(self, url):
        _NABU_VERSION_URL = "../../version"

        def get_nabu_software_verion():
            with HDF5File(url.file_path(), "r") as h5s:
                data_path = url.data_path()
                soft_version_path = os.path.normpath(
                    "/".join((data_path, _NABU_VERSION_URL))
                )
                if soft_version_path in h5s:
                    return h5py_read_dataset(h5s[soft_version_path])
                else:
                    return None

        software = software_version = ""
        nabu_version = self._decode_nabu_str(get_nabu_software_verion())
        if nabu_version is not None:
            software = "nabu"
            software_version = nabu_version
        self._softwareVersionQLE.setText("{} ({})".format(software, software_version))



# coding: utf-8
# /*##########################################################################
# Copyright (C) 2017 European Synchrotron Radiation Facility
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
#
#############################################################################*/

"""
Button of general usage.
"""

__authors__ = ["H.Payno"]
__license__ = "MIT"
__date__ = "03/10/2018"


from silx.gui import qt
from tomwer.gui import icons


class PadlockButton(qt.QPushButton):
    """Simple button to define a button with PadLock icons"""

    sigLockChanged = qt.Signal(bool)
    """signal emitted when the lock status change"""

    def __init__(self, parent):
        qt.QPushButton.__init__(self, parent)
        self._lockIcon = icons.getQIcon("locked")
        self._unlockIcon = icons.getQIcon("unlocked")
        self.setIcon(self._unlockIcon)
        self.setCheckable(True)

        # connect signals
        self.toggled.connect(self._updateDisplay)

    def setLock(self, lock):
        self.setChecked(lock)
        self._updateDisplay(lock)

    def _updateDisplay(self, checked):
        _icon = self._lockIcon if checked else self._unlockIcon
        self.setIcon(_icon)
        self.sigLockChanged.emit(checked)

    def isLocked(self):
        return self.isChecked()


class TabBrowsersButtons(qt.QWidget):
    """Simple widget containing buttons to go to 'next' or 'previous'"""

    sigPreviousReleased = qt.Signal()
    """emitted when the previous button is released"""
    sigNextReleased = qt.Signal()
    """emitted when the next button is released"""

    def __init__(self, parent=None):
        qt.QWidget.__init__(self, parent)
        self.setLayout(qt.QHBoxLayout())
        # define gui
        style = qt.QApplication.style()
        self._previousButton = qt.QPushButton("previous", self)
        previous_icon = style.standardIcon(qt.QStyle.SP_ArrowLeft)
        self._previousButton.setIcon(previous_icon)
        self.layout().addWidget(self._previousButton)

        self._nextButton = qt.QPushButton("next", self)
        next_icon = style.standardIcon(qt.QStyle.SP_ArrowRight)
        self._nextButton.setIcon(next_icon)
        self.layout().addWidget(self._nextButton)
        spacer = qt.QWidget(self)
        spacer.setSizePolicy(qt.QSizePolicy.Expanding, qt.QSizePolicy.Minimum)
        self.layout().addWidget(spacer)

        # connect signal / slot
        self._previousButton.released.connect(self._previousReleased)
        self._nextButton.released.connect(self._nextReleased)

    def _previousReleased(self, *args, **kwargs):
        self.sigPreviousReleased.emit()

    def _nextReleased(self, *args, **kwargs):
        self.sigNextReleased.emit()



# coding: utf-8
# /*##########################################################################
#
# Copyright (c) 2016-2017 European Synchrotron Radiation Facility
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
#
# ###########################################################################*/

__authors__ = ["H. Payno"]
__license__ = "MIT"
__date__ = "05/04/2019"


import os
import shutil
import tempfile
import unittest
import numpy
from tomwer.core.utils.scanutils import MockEDF, MockHDF5
from tomwer.core.process.reconstruction.darkref.params import (
    DKRFRP,
    ReduceMethod as cMethod,
)
from tomwer.core.scan.scanbase import TomwerScanBase
from ..reconstruction.darkref.darkrefs import DarkRefs
from ..reconstruction.darkref.darkrefscopy import DarkRefsCopy
from tomwer.tests.utils import UtilsTest
from tomwer.core.scan.hdf5scan import HDF5TomoScan
from tomoscan.esrf.scan.utils import get_data


class TestDarkRefIO(unittest.TestCase):
    """Test inputs and outputs types of the handler functions"""

    def setUp(self):
        self.scan_folder = tempfile.mkdtemp()

        self.scan_edf = MockEDF.mockScan(
            scanID=self.scan_folder, nRadio=10, nRecons=1, nPagRecons=4, dim=10
        )
        self.mock_hdf5 = MockHDF5(scan_path=self.scan_folder, n_proj=10, n_pag_recons=0)
        self.scan_hdf5 = self.mock_hdf5.scan

        self.recons_params = DKRFRP()

    def tearDown(self):
        shutil.rmtree(self.scan_folder)

    def testInputOutput(self):
        for scan, scan_type in zip(
            (self.scan_edf, self.scan_hdf5), ("edf scan", "hdf5 scan")
        ):
            for input_type in (dict, TomwerScanBase):
                for return_dict in (True, False):
                    with self.subTest(
                        return_dict=return_dict,
                        input_type=input_type,
                        scan_type=scan_type,
                    ):
                        input_obj = scan
                        if input_obj is dict:
                            input_obj = input_obj.to_dict()
                        process = DarkRefs(
                            inputs={
                                "dark_ref_params": self.recons_params,
                                "data": input_obj,
                                "return_dict": return_dict,
                            }
                        )
                        process.run()
                        out = process.outputs.data
                        if return_dict:
                            self.assertTrue(isinstance(out, dict))
                        else:
                            self.assertTrue(isinstance(out, TomwerScanBase))


class TestDarkRefCopyIO(unittest.TestCase):
    """Test inputs and outputs types of the handler functions"""

    def setUp(self):
        self.scan_folder = tempfile.mkdtemp()

        self.scan_edf = MockEDF.mockScan(
            scanID=self.scan_folder, nRadio=10, nRecons=1, nPagRecons=4, dim=10
        )
        self.scan_hdf5 = MockHDF5(
            scan_path=self.scan_folder, n_proj=10, n_pag_recons=0
        ).scan
        self.recons_params = DKRFRP()
        self._save_dir = tempfile.mkdtemp()

    def tearDown(self):
        shutil.rmtree(self.scan_folder)

    def testInputOutput(self):
        for scan, scan_type in zip(
            (self.scan_edf, self.scan_hdf5), ("edf scan", "hdf5 scan")
        ):
            for input_type in (dict, TomwerScanBase):
                for return_dict in (True, False):
                    with self.subTest(
                        return_dict=return_dict,
                        input_type=input_type,
                        scan_type=scan_type,
                    ):
                        input_obj = scan
                        if input_obj is dict:
                            input_obj = input_obj.to_dict()
                        dkrf_process = DarkRefsCopy(
                            inputs={
                                "dark_ref_params": self.recons_params,
                                "data": input_obj,
                                "return_dict": return_dict,
                                "save_dir": self._save_dir,
                            }
                        )
                        dkrf_process.run()
                        out = dkrf_process.outputs.data
                        if return_dict:
                            self.assertTrue(isinstance(out, dict))
                        else:
                            self.assertTrue(isinstance(out, TomwerScanBase))


class TestDarkRefEdf(unittest.TestCase):
    """
    Test dark ref is correctly processing with h5 refs
    """

    def setUp(self) -> None:
        self.scan_folder = tempfile.mkdtemp()
        self.dark_data = numpy.array(  # pylint: disable=E1121
            numpy.random.random(4 * 200 * 200) * 100,
            dtype=numpy.uint16,
        ).reshape(4, 200, 200)
        self.flat_data = numpy.array(  # pylint: disable=E1121
            numpy.random.random(4 * 200 * 200) * 100,
            dtype=numpy.uint32,
        ).reshape(4, 200, 200)
        self.scan = MockEDF.mockScan(
            scanID=self.scan_folder,
            nRadio=10,
            nRecons=1,
            nPagRecons=4,
            dim=200,
            start_dark=True,
            start_flat=True,
            start_dark_data=self.dark_data,
            start_flat_data=self.flat_data,
        )
        self.recons_params = DKRFRP()
        self.recons_params.overwrite_dark = True
        self.recons_params.overwrite_ref = True
        self.assertFalse(os.path.exists(self.scan.process_file))

    def tearDown(self) -> None:
        shutil.rmtree(self.scan_folder)

    def testDark(self):
        """
        Test darks are computed when only dark are requested
        """
        method_to_test = (
            cMethod.MEDIAN,
            cMethod.MEAN,
            cMethod.FIRST,
            cMethod.LAST,
        )
        th_results = (
            numpy.median(self.dark_data, axis=0),
            numpy.mean(self.dark_data, axis=0),
            self.dark_data[0],
            self.dark_data[-1],
        )
        for method, th_res in zip(method_to_test, th_results):
            with self.subTest(method=method):
                if os.path.exists(self.scan.process_file):
                    os.remove(self.scan.process_file)
                process = DarkRefs(
                    inputs={
                        "dark_ref_params": self.recons_params,
                        "data": self.scan,
                    }
                )

                self.recons_params.dark_calc_method = method
                self.recons_params.flat_calc_method = cMethod.NONE
                process.run()
                numpy.testing.assert_array_almost_equal(
                    self.scan.reduced_darks[0], th_res.astype(numpy.uint16)
                )

    def testFlat(self):
        """
        Test flats are computed when only flat are requested
        """
        method_to_test = (
            cMethod.MEDIAN,
            cMethod.MEAN,
            cMethod.FIRST,
            cMethod.LAST,
        )
        th_results = (
            numpy.median(self.flat_data, axis=0),
            numpy.mean(self.flat_data, axis=0),
            self.flat_data[0],
            self.flat_data[-1],
        )
        for method, th_res in zip(method_to_test, th_results):
            with self.subTest(method=method):
                if os.path.exists(self.scan.process_file):
                    os.remove(self.scan.process_file)
                self.recons_params.dark_calc_method = cMethod.NONE
                self.recons_params.flat_calc_method = method
                process = DarkRefs(
                    inputs={
                        "dark_ref_params": self.recons_params,
                        "data": self.scan,
                    }
                )
                process.run()
                numpy.testing.assert_array_almost_equal(
                    self.scan.reduced_flats[0], th_res.astype(numpy.uint16)
                )


class TestDarkRefNx(unittest.TestCase):
    """
    Test dark ref is correctly processing with h5 refs
    """

    def setUp(self) -> None:
        super().setUp()
        dataset_name = "frm_edftomomill_twoentries.nx"
        self.scan_folder = tempfile.mkdtemp()
        self._file_path = os.path.join(self.scan_folder, dataset_name)
        shutil.copyfile(
            src=UtilsTest.getH5Dataset(folderID=dataset_name), dst=self._file_path
        )
        self.scan = HDF5TomoScan(scan=self._file_path, entry="entry0000")
        self.assertFalse(os.path.exists(self.scan.process_file))
        self.recons_params = DKRFRP()
        self.recons_params.overwrite_dark = True
        self.recons_params.overwrite_ref = True

    def tearDown(self) -> None:
        shutil.rmtree(self.scan_folder)
        super().tearDown()

    def testDark(self):
        """
        Test darks are computed when only dark are requested
        """
        darks = self.scan.darks
        self.assertEqual(len(darks), 1)

        method_to_test = (cMethod.MEAN, cMethod.MEDIAN, cMethod.FIRST, cMethod.LAST)
        for method in method_to_test:
            with self.subTest(method=method):
                self.recons_params.dark_calc_method = method
                self.recons_params.flat_calc_method = cMethod.NONE
                process = DarkRefs(
                    inputs={
                        "dark_ref_params": self.recons_params,
                        "data": self.scan,
                    }
                )
                process.run()
                flats = self.scan.load_reduced_flats()
                self.assertEqual(len(flats), 0)
                darks = self.scan.load_reduced_darks()
                self.assertEqual(len(darks), 1)

    def testFlat(self):
        """
        Test flats are computed when only flat are requested
        """
        flats = self.scan.flats
        self.assertEqual(len(flats), 42)
        url_flat_serie_1 = [flats[index] for index in range(1, 22)]
        url_flat_serie_2 = [flats[index] for index in range(1521, 1542)]
        assert len(url_flat_serie_1) == 21
        assert len(url_flat_serie_2) == 21
        data_flat_serie_1 = [get_data(url) for url in url_flat_serie_1]
        data_flat_serie_2 = [get_data(url) for url in url_flat_serie_2]

        self.recons_params.overwrite_dark = True
        self.recons_params.overwrite_ref = True

        method_to_test = (cMethod.MEAN, cMethod.MEDIAN, cMethod.FIRST, cMethod.LAST)
        for method in method_to_test:
            with self.subTest(method=method):
                if method is cMethod.MEDIAN:
                    expected_res_s1 = numpy.median(data_flat_serie_1, axis=0)
                    expected_res_s2 = numpy.median(data_flat_serie_2, axis=0)
                elif method is cMethod.MEAN:
                    expected_res_s1 = numpy.mean(data_flat_serie_1, axis=0)
                    expected_res_s2 = numpy.mean(data_flat_serie_2, axis=0)
                elif method is cMethod.FIRST:
                    expected_res_s1 = data_flat_serie_1[0]
                    expected_res_s2 = data_flat_serie_2[0]
                elif method is cMethod.LAST:
                    expected_res_s1 = data_flat_serie_1[-1]
                    expected_res_s2 = data_flat_serie_2[-1]
                else:
                    raise ValueError("method not managed")

                self.recons_params.dark_calc_method = cMethod.NONE
                self.recons_params.flat_calc_method = method
                process = DarkRefs(
                    inputs={
                        "dark_ref_params": self.recons_params,
                        "data": self.scan,
                    }
                )
                process.run()

                darks = self.scan.load_reduced_darks()
                self.assertEqual(len(darks), 0)
                flats = self.scan.load_reduced_flats()
                self.assertEqual(len(flats), 2)
                self.assertTrue(1 in flats)
                self.assertTrue(1521 in flats)
                self.assertTrue(numpy.allclose(flats[1], expected_res_s1))
                self.assertTrue(numpy.allclose(flats[1521], expected_res_s2))

    def testDarkAndFlat(self):
        """
        Test darks and flats are computed when both are requested
        """
        flats = self.scan.flats
        self.assertEqual(len(flats), 42)
        url_flat_serie_1 = [flats[index] for index in range(1, 22)]
        url_flat_serie_2 = [flats[index] for index in range(1521, 1542)]
        assert len(url_flat_serie_1) == 21
        assert len(url_flat_serie_2) == 21
        data_flat_serie_1 = [get_data(url) for url in url_flat_serie_1]
        data_flat_serie_2 = [get_data(url) for url in url_flat_serie_2]

        self.recons_params.dark_calc_method = cMethod.MEAN
        self.recons_params.flat_calc_method = cMethod.MEDIAN
        darks = self.scan.darks
        self.assertEqual(len(darks), 1)
        dark_data = get_data(list(darks.values())[0])

        expected_flats_s1 = numpy.median(data_flat_serie_1, axis=0)
        expected_flats_s2 = numpy.median(data_flat_serie_2, axis=0)
        process = DarkRefs(
            inputs={
                "dark_ref_params": self.recons_params,
                "data": self.scan,
            }
        )
        process.run()
        darks = self.scan.load_reduced_darks()
        flats = self.scan.load_reduced_flats()
        self.assertEqual(len(darks), 1)
        self.assertEqual(len(flats), 2)
        self.assertTrue(0 in darks)
        self.assertTrue(1 in flats)
        self.assertTrue(1521 in flats)
        self.assertTrue(numpy.allclose(flats[1], expected_flats_s1))
        self.assertTrue(numpy.allclose(flats[1521], expected_flats_s2))
        self.assertTrue(numpy.allclose(darks[0], dark_data))



# coding: utf-8
# /*##########################################################################
#
# Copyright (c) 2016-2017 European Synchrotron Radiation Facility
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
#
# ###########################################################################*/

__authors__ = ["H. Payno"]
__license__ = "MIT"
__date__ = "04/11/2020"


from tomwer.core.scan.scanbase import TomwerScanBase
from tomwer.core.scan.hdf5scan import HDF5TomoScan
from tomwer.core.utils.scanutils import data_identifier_to_scan
from nxtomomill.utils import add_dark_flat_nx_file
from tomwer.core.process.task import Task
from tomwer.core.scan.scanfactory import ScanFactory
from silx.io.url import DataUrl
import nxtomomill.version


def apply_dark_flat_patch(scan: HDF5TomoScan, config: dict) -> TomwerScanBase:
    """

    :param scan:
    :param config:
    :return:
    """
    if not isinstance(scan, HDF5TomoScan):
        raise ValueError(
            "Dark and flat patch only manage HDF5TomoScan and "
            "not {}".format(type(scan))
        )
    if config is None:
        return scan
    for param in ("darks_start", "darks_end", "flats_start", "flats_end"):
        if param not in config:
            config[param] = None

    add_dark_flat_nx_file(
        file_path=scan.master_file,
        entry=scan.entry,
        **config,
    )
    return scan


class DarkFlatPatch(Task, input_names=("data",), output_names=("data",)):
    """
    Patch an existing NXtomo calling nxtomomill
    """

    def run(self):
        scan = data_identifier_to_scan(self.inputs.data)
        if type(scan) is dict:
            scan = ScanFactory.create_scan_object_frm_dict(scan)
        else:
            scan = scan
        if scan is None:
            return
        if not isinstance(scan, TomwerScanBase):
            raise TypeError(
                "scan is expected to be a dict or an instance "
                "of TomwerScanBase. Not {}".format(type(scan))
            )
        if not isinstance(scan, HDF5TomoScan):
            raise ValueError(
                "input type of {}: {} is not managed" "".format(scan, type(scan))
            )

        config = self.get_configuration()
        apply_dark_flat_patch(scan=scan, config=config)
        keys = config.keys()
        for key in keys:
            value = config[key]
            if isinstance(value, DataUrl):
                config[key] = value.path()

        with scan.acquire_process_file_lock():
            self.register_process(
                process_file=scan.process_file,
                entry=scan.entry,
                configuration=config,
                results={},
                process_index=scan.pop_process_index(),
                overwrite=True,
            )
        self.outputs.data = scan

    @staticmethod
    def program_name():
        return "nxtomomill.utils.change_image_key_control"

    @staticmethod
    def program_version():
        return nxtomomill.version.version

    @staticmethod
    def definition():
        return "Apply patch for dark and references on a scan (TomwerScanBase)"



# coding: utf-8
# /*##########################################################################
# Copyright (C) 2017 European Synchrotron Radiation Facility
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
#
#############################################################################*/

"""
This module is used to define a set of folders to be emitted to the next box.
"""

__authors__ = ["H.Payno"]
__license__ = "MIT"
__date__ = "05/07/2017"


from tomwer.core.process.task import Task
from tomwer.core.utils.scanutils import data_identifier_to_scan
import logging

logger = logging.getLogger(__name__)


class _ScanListPlaceHolder(
    Task, optional_input_names=("data",), output_names=("data",)
):
    """For now data can only be a single element and not a list.
    This must be looked at.
    Also when part of an ewoks graph 'data' is mandatory which is not the class
    when part of a orange workflow. Those can be added interactively"""

    def run(self):
        self.outputs.data = data_identifier_to_scan(self.inputs.data)



# coding: utf-8
# /*##########################################################################
#
# Copyright (c) 2016-2017 European Synchrotron Radiation Facility
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
#
# ###########################################################################*/

__authors__ = ["H. Payno"]
__license__ = "MIT"
__date__ = "01/06/2018"


import os
from silx.gui import qt
import logging
from tomwer.core.process.reconstruction.lamino.tofu import getDark, getFlats, FFCWhen
from tomwer.core.utils import getDARK_N, getDim1Dim2
from tomwer.core.utils import getParametersFromParOrInfo
from tomwer.core.utils import getTomo_N, getPixelSize, getClosestEnergy
from tomwer.core.utils.char import PSI_CHAR, THETA_CHAR
from tomwer.gui import icons
from tomwer.gui.utils.illustrations import _IllustrationWidget, _IllustrationDialog
from tomwer.gui.utils.unitsystem import MetricEntry
from .TofuOptionLoader import _TofuOptionLoader, _getterSetter
from .misc import _AngleWidget, PadlockButton

_logger = logging.getLogger(__name__)


class InputProjectionsScrollArea(qt.QScrollArea):
    def __init__(self, parent):
        qt.QScrollArea.__init__(self, parent)
        self.widget = InputProjectionsWidget(parent=self)
        self.setWidget(self.widget)
        self.setWidgetResizable(True)

        # expose API
        self.loadFromScan = self.widget.loadFromScan
        self.setNumber = self.widget.setNumber
        self.getNumber = self.widget.getNumber
        self.getPixelSize = self.widget.getPixelSize
        self.setPixelSize = self.widget.setPixelSize
        self.pixelSizeIsLocked = self.widget.pixelSizeIsLocked
        self.getParameters = self.widget.getParameters
        self.loadFromScan = self.widget.loadFromScan
        self.setParameters = self.widget.setParameters
        self.rotationAngle = self.widget.rotationAngle
        self.centeringWidget = self.widget.centeringWidget
        self.getWhenApplyFFC = self.widget.getWhenApplyFFC
        self.setWhenApplyFFC = self.widget.setWhenApplyFFC
        self.isHalfAcquisition = self.widget.isHalfAcquisition
        self.setHalfAcquisition = self.widget.setHalfAcquisition
        self.setBlend = self.widget.setBlend
        self.setAdjustMean = self.widget.setAdjustMean


class InputProjectionsWidget(_TofuOptionLoader, qt.QWidget):
    """TabWidget containing all the information relative to the
    flat field correction"""

    def __init__(self, parent):
        qt.QWidget.__init__(self, parent)
        self.setLayout(qt.QGridLayout())

        self.rotationAngle = RotationAngleGroupBox(parent=self)
        self.layout().addWidget(self.rotationAngle, 0, 0)

        self.centeringWidget = CenteringTofuGroupBox(parent=self)
        self.layout().addWidget(self.centeringWidget, 1, 0)

        self._numberOfFiles = NbFilesWidget(parent=self)
        self._numberOfFiles.layout().setContentsMargins(6, 2, 2, 2)
        self.layout().addWidget(self._numberOfFiles, 2, 0)

        # pixel size
        self._pixelWidget = qt.QWidget(parent=self)
        self._pixelWidget.setLayout(qt.QHBoxLayout())
        self._pixelWidget.layout().setContentsMargins(6, 2, 2, 2)

        self._pixelSize = MetricEntry(name="pixel size", parent=self, default_unit="m")
        self._pixelWidget.layout().addWidget(self._pixelSize._label)
        self._pixelWidget.layout().addWidget(self._pixelSize._qlePixelSize)
        self._pixelWidget.layout().addWidget(self._pixelSize._qcbUnit)
        self._pixelSizeLockButton = PadlockButton(parent=self)
        self._pixelWidget.layout().addWidget(self._pixelSizeLockButton)
        self.layout().addWidget(self._pixelWidget, 3, 0)

        # half  & options
        self._halfAcqWidget = qt.QWidget(parent=self)
        self._halfAcqWidget.setLayout(qt.QHBoxLayout())
        self._halfAcqWidget.layout().setContentsMargins(6, 2, 2, 2)
        self._halfAcqCB = qt.QCheckBox("half acquisition", parent=self)
        self._halfAcqWidget.layout().addWidget(self._halfAcqCB)
        self._blendCB = qt.QCheckBox("blend", parent=self)
        self._blendCB.setVisible(False)
        self._halfAcqWidget.layout().addWidget(self._blendCB)
        self._adjustMeanCB = qt.QCheckBox("adjust mean", parent=self)
        self._adjustMeanCB.setVisible(False)
        self._halfAcqWidget.layout().addWidget(self._adjustMeanCB)
        self.layout().addWidget(self._halfAcqWidget, 4, 0)

        # flat field correction
        self._ffcWidget = _FFCConfigWidget(parent=self)
        self.layout().addWidget(self._ffcWidget, 5, 0, 2, 2)

        self._phaseRetrieval = PhaseRetrievalWidget(parent=self)
        self._phaseRetrieval.setCheckable(True)
        self._phaseRetrieval.setChecked(False)
        self.layout().addWidget(self._phaseRetrieval, 7, 0)

        spacer = qt.QWidget(self)
        spacer.setSizePolicy(qt.QSizePolicy.Minimum, qt.QSizePolicy.Expanding)
        self.layout().addWidget(spacer)

        self._ffcWidget._absortivityActivated.connect(self._phaseRetrieval.setUnchecked)
        self._phaseRetrieval.activated.connect(self._ffcWidget.uncheckedAbsorptivity)

        # expose API
        self.setNumber = self._numberOfFiles.setNumberOfFiles
        self.getNumber = self._numberOfFiles.getNumberOfFiles
        self.getPixelSize = self._pixelSize.getValue
        self.setPixelSize = self._pixelSize.setValue
        self.pixelSizeIsLocked = self._pixelSizeLockButton.isLocked
        self.getWhenApplyFFC = self._ffcWidget.getWhenApplyFFC
        self.setWhenApplyFFC = self._ffcWidget.setWhenApplyFFC
        self.isHalfAcquisition = self._halfAcqCB.isChecked
        self.setHalfAcquisition = self._halfAcqCB.setChecked
        self.isBlend = self._blendCB.isChecked
        self.setBlend = self._blendCB.setChecked
        self.isAdjustMean = self._adjustMeanCB.isChecked
        self.setAdjustMean = self._adjustMeanCB.setChecked

        options = {
            "ffc-when": _getterSetter(
                getter=self.getWhenApplyFFC, setter=self.setWhenApplyFFC
            ),
            "number": _getterSetter(getter=self.getNumber, setter=self.setNumber),
            "reduction-mode": _getterSetter(
                getter=self._ffcWidget.getMethod, setter=self._ffcWidget.setMethod
            ),
            "darks": _getterSetter(
                getter=self._ffcWidget.getDarks, setter=self._ffcWidget.setDarks
            ),
            "flats": _getterSetter(
                getter=self._ffcWidget.getFlats, setter=self._ffcWidget.setFlats
            ),
            "flats2": _getterSetter(
                getter=self._ffcWidget.getSecondFlats,
                setter=self._ffcWidget.setSecondFlats,
            ),
            "absorptivity": _getterSetter(
                getter=self._ffcWidget.requireAbsorptivity,
                setter=self._ffcWidget.setAbsortivity,
            ),
            "fix-nan-and-inf": _getterSetter(
                getter=self._ffcWidget.removeNanAndInf,
                setter=self._ffcWidget.setNanAndInf,
            ),
            "dark-scale": _getterSetter(
                getter=self._ffcWidget.getDarkScale, setter=self._ffcWidget.setDarkScale
            ),
            "axis-angle-x": _getterSetter(
                getter=self.rotationAngle._laminoAngle.getAngle,
                setter=self.rotationAngle._laminoAngle.setAngle,
            ),
            "axis-angle-y": _getterSetter(
                getter=self.rotationAngle._axisAngleY.getAngle,
                setter=self.rotationAngle._axisAngleY.setAngle,
            ),
            "axis-angle-z": _getterSetter(
                getter=self.rotationAngle._axisAngleZ.getAngle,
                setter=self.rotationAngle._axisAngleZ.setAngle,
            ),
            "center-position-x": _getterSetter(
                getter=self.centeringWidget.getXCenter,
                setter=self.centeringWidget.setXCenter,
            ),
            "center-position-z": _getterSetter(
                getter=self.centeringWidget.getZCenter,
                setter=self.centeringWidget.setZCenter,
            ),
            "overall-angle": _getterSetter(
                getter=self.rotationAngle._overallAngle.getAngle,
                setter=self.rotationAngle._setOverallAngleI,
            ),
            "pixel-size": _getterSetter(
                getter=self.getPixelSize, setter=self.setPixelSize
            ),
            "half-acquisition": _getterSetter(
                getter=self.isHalfAcquisition, setter=self._halfAcqCB.setChecked
            ),
            "blend": _getterSetter(getter=self.isBlend, setter=self.setBlend),
            "adjust-mean": _getterSetter(
                getter=self.isAdjustMean, setter=self.setAdjustMean
            ),
        }
        _TofuOptionLoader.__init__(self, options=options, childs=[self._phaseRetrieval])

        # connect
        self._halfAcqCB.toggled.connect(self._updateFFCOptions)

    def _updateFFCOptions(self):
        # if we want the half acquisition, for now we force the ffc to be done
        # in preprocessing
        self._ffcWidget.forcePreprocessing(self._halfAcqCB.isChecked())
        self._blendCB.setVisible(self._halfAcqCB.isChecked())
        self._adjustMeanCB.setVisible(self._halfAcqCB.isChecked())

    def loadFromScan(self, scanID):
        if scanID is None:
            return
        try:
            nFile = getTomo_N(scanID)
            self.setNumber(nFile)
        except Exception as error:
            _logger.error(error)

        self.rotationAngle.loadFromScan(scanID=scanID)
        self.centeringWidget.loadFromScan(scanID=scanID)
        self._ffcWidget.loadFromScan(scanID=scanID)
        self._phaseRetrieval.loadFromScan(scanID=scanID)

        pixelSize = getPixelSize(scanID)
        if pixelSize is not None and self.pixelSizeIsLocked() is False:
            self.setPixelSize(float(pixelSize))


class _FFCConfigWidget(qt.QGroupBox):
    """Widget dedicated to the parameters used for the flat field correction"""

    _absortivityActivated = qt.Signal()
    """Signal emitted when the absorptivity is activated"""

    def __init__(self, parent):
        super(_FFCConfigWidget, self).__init__("flat field correction", parent)
        self.setLayout(qt.QGridLayout())

        self._whenGrp = _FFCWhenWidget(parent=self)
        self.layout().addWidget(qt.QLabel("when"), 0, 0)
        self.layout().addWidget(self._whenGrp, 0, 1)

        self.layout().addWidget(qt.QLabel("method"), 1, 0)
        self._method = qt.QComboBox(parent=self)
        self._method.addItem("median")
        self._method.addItem("Average")
        self.layout().addWidget(self._method, 1, 1)

        self.__absortivityLabel = qt.QLabel("absorptivity (*)")
        self.__absortivityLabel.setToolTip(
            "Activating absortivity will " "deactivate phase retrieval"
        )
        self.layout().addWidget(self.__absortivityLabel, 2, 0)

        self._absortivity = qt.QCheckBox(parent=self)
        self.layout().addWidget(self._absortivity, 2, 1)
        self._absortivity.setChecked(True)

        self.layout().addWidget(qt.QLabel("remove Nan and Inf"), 3, 0)
        self._removeNanInf = qt.QCheckBox(parent=self)
        self.layout().addWidget(self._removeNanInf, 3, 1)
        self._removeNanInf.setChecked(True)

        self.layout().addWidget(qt.QLabel("dark scale"), 4, 0)
        self._darkScaleSB = qt.QDoubleSpinBox(parent=self)
        self._darkScaleSB.setMinimum(0.0)
        self._darkScaleSB.setValue(1.0)
        self._darkScaleSB.setSingleStep(0.1)
        self.layout().addWidget(self._darkScaleSB, 4, 1)

        self.layout().addWidget(qt.QLabel("darks"), 5, 0)
        self._darksLE = qt.QLineEdit("", parent=self)
        self.layout().addWidget(self._darksLE, 5, 1)
        self.layout().addWidget(qt.QLabel("flats"), 6, 0)
        self._flatsLE = qt.QLineEdit("", parent=self)
        self.layout().addWidget(self._flatsLE, 6, 1)

        self.layout().addWidget(qt.QLabel("second flats"), 7, 0)
        self._secondFlatsLE = qt.QLineEdit("", parent=self)
        self.layout().addWidget(self._secondFlatsLE, 7, 1)

        self._absortivity.toggled.connect(self._absortivityStatusChanged)

        # expose API
        self.getWhenApplyFFC = self._whenGrp.getWhen
        self.setWhenApplyFFC = self._whenGrp.setWhen
        self.forcePreprocessing = self._whenGrp.forcePreprocessing

    def _absortivityStatusChanged(self, checked):
        if checked is True:
            self._absortivityActivated.emit()

    def uncheckedAbsorptivity(self):
        self._absortivity.setChecked(False)

    def getMethod(self):
        return self._method.currentText()

    def setMethod(self, method):
        _method = method
        if _method.lower == "median":
            _method = "median"
        if _method.title() == "Average":
            _method = "Average"
        index = self._method.findText(_method)
        if index < 0:
            _logger.warning("fail to set method %s, unrecognized method: ", _method)
        else:
            self._method.setCurrentIndex(index)

    def setAbsortivity(self, value=True):
        _value = value
        if _value == "":
            _value = True
        self._absortivity.setChecked(_value)

    def setNanAndInf(self, value=True):
        _value = value
        if _value == "":
            _value = True
        self._removeNanInf.setChecked(_value)

    def requireAbsorptivity(self):
        return self._absortivity.isChecked()

    def removeNanAndInf(self):
        return self._removeNanInf.isChecked()

    def getDarks(self):
        if self._darksLE.text() == "":
            return None
        else:
            return self._darksLE.text()

    def setDarks(self, val):
        self._darksLE.setText(val)

    def getFlats(self):
        if self._flatsLE.text() == "":
            return None
        else:
            return self._flatsLE.text()

    def setFlats(self, val):
        self._flatsLE.setText(val)

    def getSecondFlats(self):
        if self._secondFlatsLE.text() == "":
            return None
        else:
            return self._secondFlatsLE.text()

    def setSecondFlats(self, val):
        self._secondFlatsLE.setText(val)

    def hasDarks(self):
        return self._darksLE.text() != ""

    def getDarkScale(self):
        if self.hasDarks() is False:
            return None
        else:
            return float(self._darkScaleSB.value())

    def setDarkScale(self, value):
        if value is not None:
            self._darkScaleSB.setValue(float(value))

    def resetDarkScale(self):
        self._darkScaleSB.setValue(1.0)

    def loadFromScan(self, scanID):
        try:
            flats, secondFlat = getFlats(scan=scanID)
            self.setFlats(flats or "")
            self.setSecondFlats(secondFlat or "")
            darkN = getDARK_N(scanID)
            if darkN is not None:
                self._darkScaleSB.setValue(1.0 / int(darkN))
            darkFile = getDark(scan=scanID)
            if darkFile is not None:
                self.setDarks(darkFile)
                # in this case dark is already normalized, set dark scale to 1.0
                if os.path.basename(darkFile) in ("dark.edf", "darkHST.edf"):
                    self.resetDarkScale()
        except Exception as error:
            _logger.error(error)


class _FFCWhenWidget(qt.QWidget):
    def __init__(self, parent):
        qt.QWidget.__init__(self, parent)
        self.setLayout(qt.QHBoxLayout())
        self.layout().setContentsMargins(0, 0, 0, 0)
        self._group = qt.QButtonGroup(parent=self)

        name = FFCWhen.on_the_fly.name.replace("_", " ")
        self._onTheFlyQRB = qt.QRadioButton(name, self)
        self._onTheFlyQRB.setToolTip(
            "process flat field correction during "
            "reconstruction. Avoid multiple io read."
        )
        self.layout().addWidget(self._onTheFlyQRB)
        self._group.addButton(self._onTheFlyQRB)

        name = FFCWhen.preprocessing.name.replace("_", " ")
        self._preProcQRB = qt.QRadioButton(name, self)
        self._preProcQRB.setToolTip(
            "make flat field on a preprocessing step. "
            "Can be interesting if you intend to test "
            "several reconstruction parameters"
        )
        self.layout().addWidget(self._preProcQRB)
        self._group.addButton(self._preProcQRB)

        # setting: default on the fly
        self._onTheFlyQRB.setChecked(True)

    def getWhen(self):
        if self._onTheFlyQRB.isChecked():
            return FFCWhen.on_the_fly
        elif self._preProcQRB.isChecked():
            return FFCWhen.preprocessing
        else:
            raise ValueError('No "when" defined')

    def setWhen(self, name):
        if isinstance(name, FFCWhen):
            if name is FFCWhen.on_the_fly:
                self._onTheFlyQRB.setChecked(True)
            elif name is FFCWhen.preprocessing:
                self._preProcQRB.setChecked(True)
            else:
                raise ValueError("FFCWhen type is not managed")
        elif type(name) is str:
            try:
                _name = name.split(".")[-1]
                #  name.split('.')[-1] because name can be given as FFCWhen.on_the_fly...
                _when_instance = getattr(FFCWhen, _name.lower().replace(" ", "_"))
            except Exception:
                _logger.error(f"fail to determine 'FFCWhen' value from {name}")
            else:
                self.setWhen(_when_instance)
        else:
            raise ValueError("name should be a str or a FFCWhen")

    def forcePreprocessing(self, force_):
        old = self.blockSignals(True)
        if force_:
            self._preProcQRB.setChecked(True)
        self.setEnabled(not force_)
        self.blockSignals(old)


class CenteringTofuGroupBox(qt.QGroupBox):
    """
    Widget used to display centering information
    """

    def __init__(self, parent):
        qt.QGroupBox.__init__(self, "centering", parent)
        self.setLayout(qt.QGridLayout())
        self.layout().addWidget(qt.QLabel("x center", parent=self), 0, 0)
        self._xCenterLE = qt.QLineEdit("0", parent=self)
        validator = qt.QDoubleValidator(parent=self)
        self._xCenterLE.setValidator(validator)
        self.layout().addWidget(self._xCenterLE, 0, 1)

        self.layout().addWidget(qt.QLabel("y center", parent=self), 1, 0)
        self._yCenterLE = qt.QLineEdit("0", parent=self)
        self._yCenterLE.setValidator(validator)
        self.layout().addWidget(self._yCenterLE, 1, 1)

        self._lockButton = PadlockButton(parent=self)
        self._lockButton.setSizePolicy(
            qt.QSizePolicy.Expanding, qt.QSizePolicy.Expanding
        )
        self.layout().addWidget(self._lockButton, 0, 2, 2, 2)

        # expose API
        self.isLocked = self._lockButton.isLocked

    def getXCenter(self):
        return float(self._xCenterLE.text())

    def setXCenter(self, value):
        self._xCenterLE.setText(str(value))

    def getZCenter(self):
        return float(self._yCenterLE.text())

    def setZCenter(self, value):
        self._yCenterLE.setText(str(value))

    def loadFromScan(self, scanID):
        if self.isLocked() is False:
            # can be set from Dim 1 and Dim 2
            # question: here to avoid interpolation we simply make an integer
            # division which give an integer
            dim1, dim2 = getDim1Dim2(scanID)
            for dim, lineEdit in zip((dim1, dim2), (self._xCenterLE, self._yCenterLE)):
                if dim is not None:
                    lineEdit.setText(str(int(dim / 2)))

            # try to set x center from the .par file
            try:
                parFile = os.path.join(scanID, os.path.basename(scanID) + ".par")
                if os.path.exists(parFile) is True:
                    params = getParametersFromParOrInfo(parFile)
                    rot_axis_pos = None
                    for k in ("ROTATION_AXIS_POS", "ROTATION_AXIS_POSITION"):
                        if k.lower() in params:
                            rot_axis_pos = params[k.lower()]

                    if rot_axis_pos is not None:
                        self._xCenterLE.setText(str(rot_axis_pos))
            except Exception as error:
                _logger.error(error)


class NbFilesWidget(qt.QWidget):
    """
    Widget to display the number of file (projection) to be send to tofu
    """

    def __init__(self, parent=None):
        qt.QWidget.__init__(self, parent=parent)
        self.setLayout(qt.QGridLayout())
        self.layout().addWidget(qt.QLabel("number of files", parent=self), 0, 0)
        self._nFileLE = qt.QLineEdit("0", parent=self)
        validator = qt.QIntValidator()
        validator.setBottom(0)
        self._nFileLE.setValidator(validator)
        self.layout().addWidget(self._nFileLE, 0, 1)

    def setNumberOfFiles(self, n):
        self._nFileLE.setText(str(n))

    def getNumberOfFiles(self):
        if self._nFileLE.text().isdigit():
            return int(self._nFileLE.text())
        else:
            return 0


class AngleIllustrations(qt.QWidget):
    def __init__(self, parent):
        qt.QWidget.__init__(self, parent)
        self.setLayout(qt.QVBoxLayout())
        self.setContentsMargins(0, 0, 0, 0)
        self._infoLabel = qt.QLabel(parent=self)
        icon = icons.getQIcon("information")
        self._infoLabel.setPixmap(icon.pixmap(qt.QSize(32, 32)))
        self.layout().addWidget(self._infoLabel)

        self.illustrationCB = qt.QComboBox(parent=self)
        self.illustrationCB.addItem("lamino angle")
        self.illustrationCB.addItem(PSI_CHAR + " angle")
        self.layout().addWidget(self.illustrationCB)

        self._illustrations = qt.QWidget(parent=self)
        self._illustrations.setLayout(qt.QHBoxLayout())
        self.layout().addWidget(self._illustrations)
        self._laminoAngleIllustration = _IllustrationWidget(
            parent=self, img="lamino_angle"
        )
        self._psiAngleIllustration = _IllustrationWidget(parent=self, img="psi_angle")
        self._illustrations.setMinimumSize(qt.QSize(200, 200))
        self._illustrations.setSizePolicy(
            qt.QSizePolicy.Expanding, qt.QSizePolicy.Expanding
        )

        self._illustrations.setContentsMargins(0, 0, 0, 0)
        self._illustrations.layout().setSpacing(0)
        self._illustrations.layout().addWidget(self._laminoAngleIllustration)
        self._illustrations.layout().addWidget(self._psiAngleIllustration)
        self.layout().addWidget(self._illustrations)
        self.setActiveIllustration("lamino angle")
        self.illustrationCB.currentIndexChanged[str].connect(self.setActiveIllustration)

    def setActiveIllustration(self, name):
        assert name in ("lamino angle", PSI_CHAR + " angle", THETA_CHAR + " angle")
        if name == THETA_CHAR + " angle":
            _logger.info("No illustration of the delta angle for the moment")
            return
        self.blockSignals(True)
        index = self.illustrationCB.findText(name)
        assert index >= 0
        self.illustrationCB.setCurrentIndex(index)

        self._laminoAngleIllustration.setVisible(name == "lamino angle")
        self._psiAngleIllustration.setVisible(name == PSI_CHAR + " angle")

        self.blockSignals(False)


class RotationAngleGroupBox(qt.QGroupBox):
    def __init__(self, parent):
        qt.QGroupBox.__init__(self, "angles", parent)
        self.setLayout(qt.QHBoxLayout())

        self.layout().setContentsMargins(0, 0, 0, 0)
        self.layout().setSpacing(0)

        self._angles = qt.QWidget(parent=self)
        self._angles.setLayout(qt.QVBoxLayout())
        self._angles.setContentsMargins(0, 0, 0, 0)
        self.layout().addWidget(self._angles)

        self._laminoAngle = _AngleWidget(
            parent=self._angles,
            name="rotation axis tilt",
            defaultVal=90,
            lockable=True,
            information=self._showLaminoAngleInfo,
        )
        self._angles.layout().addWidget(self._laminoAngle)
        self._axisAngleY = _AngleWidget(
            parent=self._angles,
            name=(PSI_CHAR + " angle"),
            lockable=True,
            information=self._showPsiAngleInfo,
        )
        self._angles.layout().addWidget(self._axisAngleY)

        self._axisAngleZ = _AngleWidget(
            parent=self._angles, name=(THETA_CHAR + " angle"), lockable=True
        )
        self._angles.layout().addWidget(self._axisAngleZ)
        self._axisAngleZ.setVisible(False)

        self._overallAngle = _AngleWidget(
            parent=self, name="overall angle", defaultVal=360, lockable=True
        )
        self._angles.layout().addWidget(self._overallAngle)

        self._spacer = qt.QWidget(parent=self._angles)
        self._spacer.setSizePolicy(qt.QSizePolicy.Minimum, qt.QSizePolicy.Expanding)
        self._angles.layout().addWidget(self._spacer)

        self._infoLamino = None
        self._infoPsi = None

        # expose some API
        self.getLaminoAngle = self._laminoAngle.getAngle
        self.getPsiAngle = self._axisAngleY.getAngle

    def _showLaminoAngleInfo(self):
        self.getInfoLaminoDialog().show()
        self.getInfoLaminoDialog().raise_()
        self.getInfoLaminoDialog().activateWindow()

    def _showPsiAngleInfo(self):
        self.getPsiLaminoDialog().show()
        self.getPsiLaminoDialog().raise_()
        self.getPsiLaminoDialog().activateWindow()

    def getInfoLaminoDialog(self):
        if self._infoLamino is None:
            self._infoLamino = _IllustrationDialog(
                parent=self, img="lamino_angle", title="information about lamino angle"
            )
        return self._infoLamino

    def getPsiLaminoDialog(self):
        if self._infoPsi is None:
            self._infoPsi = _IllustrationDialog(
                parent=self,
                img="psi_angle",
                title=" ".join(("information about", PSI_CHAR, "angle")),
            )
        return self._infoPsi

    def loadFromScan(self, scanID):
        assert type(scanID) is str
        _info_file = os.path.join(scanID, os.path.basename(scanID) + ".info")
        if os.path.isfile(_info_file) is True:
            info = getParametersFromParOrInfo(_info_file)
            if self._laminoAngle.isLocked() is False and "ctangle" in info:
                self._laminoAngle.setAngle(90.0 - float(info["ctangle"]))
            if self._overallAngle.isLocked() is False and "scanrange" in info:
                self._overallAngle.setAngle(float(info["scanrange"]))

    def _setOverallAngleI(self, val):
        """
        Set the inverse value to the lamino angle

        :param val:
        :return:
        """
        _val = float(val) * -1.0
        self._overallAngle.setAngle(_val)


class PhaseRetrievalWidget(_TofuOptionLoader, qt.QGroupBox):
    """
    Widget sued to defined the phase retrieval
    """

    activated = qt.Signal()
    """Emitted when the phase retrieval is activated"""

    METHODS = "tie", "ctf", "ctfhalfsin", "qp", "qphalfsine", "qp2"
    DEFAULT_METHOD = "tie"

    def __init__(self, parent):
        qt.QGroupBox.__init__(self, "phase retrieval (*)", parent)
        self.setToolTip("Activating phase retrieval will deactivate " "absorptivity")
        self.setLayout(qt.QGridLayout())
        # retrieval method
        self._retrievalMethodCB = qt.QComboBox(parent=self)
        for _method in self.METHODS:
            self._retrievalMethodCB.addItem(_method)
        assert self.DEFAULT_METHOD in self.METHODS
        self._retrievalMethodCB.setCurrentIndex(
            self._retrievalMethodCB.findText(self.DEFAULT_METHOD)
        )
        self.layout().addWidget(qt.QLabel("retrieval method:", self), 0, 0)
        self.layout().addWidget(self._retrievalMethodCB, 0, 1)

        # energy
        self.layout().addWidget(qt.QLabel("energy:", self), 1, 0)
        self._energy = qt.QLineEdit("65.0", parent=self)
        validator = qt.QDoubleValidator(parent=self._energy)
        self._energy.setValidator(validator)
        self.layout().addWidget(self._energy, 1, 1)
        self._energyLockButton = PadlockButton(parent=self)
        self.layout().addWidget(self._energyLockButton, 1, 2)

        # regularization rate
        self.layout().addWidget(qt.QLabel("regularization rate:", self), 3, 0)
        self._regularizationRate = qt.QLineEdit("1.5", parent=self)
        validator = qt.QDoubleValidator(parent=self._regularizationRate)
        self._regularizationRate.setValidator(validator)
        self.layout().addWidget(self._regularizationRate, 3, 1)
        self._regRateLockButton = PadlockButton(parent=self)
        self.layout().addWidget(self._regRateLockButton, 3, 2)

        # thresholding rate
        self.layout().addWidget(qt.QLabel("thresholding rate:", self), 4, 0)
        self._thresholdingRate = qt.QLineEdit("0.4", parent=self)
        validator = qt.QDoubleValidator(parent=self._thresholdingRate)
        self._thresholdingRate.setValidator(validator)
        self.layout().addWidget(self._thresholdingRate, 4, 1)

        # propagation distance
        self.layout().addWidget(qt.QLabel("propagation distance (m):", self), 5, 0)
        self._propagationDistance = qt.QLineEdit("0.08", parent=self)
        validator = qt.QDoubleValidator(parent=self._propagationDistance)
        self._propagationDistance.setValidator(validator)
        self.layout().addWidget(self._propagationDistance, 5, 1)
        self._propaDistLockButton = PadlockButton(parent=self)
        self.layout().addWidget(self._propaDistLockButton, 5, 2)

        # create some more speaking API
        self.energyIsLocked = self._energyLockButton.isLocked
        self.getRetrievalMethod = self._retrievalMethodCB.currentText
        self.propaDistIsLocked = self._propaDistLockButton.isLocked

        # signal/slot connection
        self.toggled.connect(self.__toggleCallback)

        options = {
            "retrieval-method": _getterSetter(
                getter=self._getRetrievalMethod, setter=self.setRetrievalMethod
            ),
            "energy": _getterSetter(getter=self._getEnergy, setter=self.setEnergy),
            "propagation-distance": _getterSetter(
                getter=self._getPropagationDistance, setter=self.setPropagationDistance
            ),
            "regularization-rate": _getterSetter(
                getter=self._getRegularizationRate, setter=self.setRegularizationRate
            ),
            "thresholding-rate": _getterSetter(
                getter=self._getThresholdingRate, setter=self.setThresholdingRate
            ),
        }
        _TofuOptionLoader.__init__(self, options=options)

    def loadFromScan(self, scanID):
        _info_file = os.path.join(scanID, os.path.basename(scanID) + ".info")
        if os.path.isfile(_info_file) is True:
            energy = getClosestEnergy(scanID)
            if energy is not None and self.energyIsLocked() is False:
                self.setEnergy(energy)
            info = getParametersFromParOrInfo(_info_file)
            if "distance" in info and self.propaDistIsLocked() is False:
                try:
                    dst = float(info["distance"])
                    # convert to meter
                    dst = dst / 1000.0
                    self.setPropagationDistance(dst)
                except Exception:
                    pass

    def setUnchecked(self):
        self.setChecked(False)

    def __toggleCallback(self, checked):
        if checked is True:
            self.activated.emit()

    # retrieval method
    def _getRetrievalMethod(self):
        if self.isChecked() is True:
            return self.getRetrievalMethod()
        else:
            return None

    def setRetrievalMethod(self, method):
        if method is not None:
            if method not in self.METHODS:
                _logger.error("method %s is not recognized as a retrieval " "method")
            else:
                index = self._retrievalMethodCB.find(method)
                self._retrievalMethodCB.setCurrentIndex(index)

    # energy
    def getEnergy(self):
        return float(self._energy.text())

    def _getEnergy(self):
        "Return None if not activated, otherwise the value"
        if self.isChecked() is True:
            return self.getEnergy()

    def setEnergy(self, energy):
        if energy is not None:
            self._energy.setText(str(energy))

    # propagation distance
    def getPropagationDistance(self):
        return float(self._propagationDistance.text())

    def _getPropagationDistance(self):
        if self.isChecked() is True:
            return self.getPropagationDistance()
        else:
            return None

    def setPropagationDistance(self, distance):
        if distance is not None:
            self._propagationDistance.setText(str(distance))

    # regularization rate
    def getRegularizationRate(self):
        return float(self._regularizationRate.text())

    def _getRegularizationRate(self):
        if self.isChecked() is True:
            return self.getRegularizationRate()
        else:
            return None

    def setRegularizationRate(self, rate):
        if rate is not None:
            self._regularizationRate.setText(str(rate))

    # thresholding rate
    def getThresholdingRate(self):
        return float(self._thresholdingRate.text())

    def _getThresholdingRate(self):
        if self.isChecked() is True:
            return self.getThresholdingRate()
        else:
            return None

    def setThresholdingRate(self, rate):
        if rate is not None:
            self._thresholdingRate.setText(str(rate))



# coding: utf-8
#
#    Copyright (C) 2012-2016 European Synchrotron Radiation Facility, Grenoble, France
#
#    Principal author:       Jérôme Kieffer (Jerome.Kieffer@ESRF.eu)
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
#
"bunch of utility function/static classes to handle testing environment"

__author__ = "Jérôme Kieffer"
__contact__ = "jerome.kieffer@esrf.eu"
__license__ = "MIT"
__copyright__ = "European Synchrotron Radiation Facility, Grenoble, France"
__date__ = "24/01/2018"

import os
import sys
import threading
import unittest
import logging
import shutil
import tempfile
import getpass


from argparse import ArgumentParser
from silx.resources import ExternalResources

PACKAGE = "pyFAI"
DATA_KEY = "PYFAI_DATA"

logger = logging.getLogger(__name__)

TEST_HOME = os.path.dirname(os.path.abspath(__file__))


def copy(infile, outfile):
    "link or copy file according to the OS"
    if "link" in dir(os):
        os.link(infile, outfile)
    else:
        shutil.copy(infile, outfile)


class TestContext(object):
    """
    Class providing useful stuff for preparing tests.
    """

    def __init__(self):
        self.options = None
        self.timeout = 60  # timeout in seconds for downloading images
        self.url_base = "http://ftp.edna-site.org/pyFAI/testimages"
        self.resources = ExternalResources(
            PACKAGE, timeout=self.timeout, env_key=DATA_KEY, url_base=self.url_base
        )
        self.sem = threading.Semaphore()
        self.recompiled = False
        self.reloaded = False
        self.name = PACKAGE
        self.script_dir = None
        self.installed = False

        self.download_images = self.resources.download_all
        self.getimage = self.resources.getfile
        self.low_mem = bool(os.environ.get("PYFAI_LOW_MEM"))
        self.opencl = bool(os.environ.get("PYFAI_OPENCL", True))

        self._tempdir = None

    def forceBuild(self, remove_first=True):
        """
        Force the recompilation of pyFAI

        Nonesense, kept for legacy reasons
        """
        return

    def get_options(self):
        """
        Parse the command line to analyse options ... returns options
        """
        if self.options is None:
            parser = ArgumentParser(usage="Tests for %s" % self.name)
            parser.add_argument(
                "-d",
                "--debug",
                dest="debug",
                help="run in debugging mode",
                default=False,
                action="store_true",
            )
            parser.add_argument(
                "-i",
                "--info",
                dest="info",
                help="run in more verbose mode ",
                default=False,
                action="store_true",
            )
            parser.add_argument(
                "-f",
                "--force",
                dest="force",
                help="force the build of the library",
                default=False,
                action="store_true",
            )
            parser.add_argument(
                "-r",
                "--really-force",
                dest="remove",
                help="remove existing build and force the build of the library",
                default=False,
                action="store_true",
            )
            parser.add_argument(dest="args", type=str, nargs="*")
            self.options = parser.parse_args([])
        return self.options

    def get_test_env(self):
        """
        Returns an associated environment with a working project.
        """
        env = dict((str(k), str(v)) for k, v in os.environ.items())
        env["PYTHONPATH"] = os.pathsep.join(sys.path)
        return env

    def script_path(self, script_name, module_name):
        """Returns the script path according to it's location"""
        if self.installed:
            script = self.get_installed_script_path(script_name)
        else:
            import importlib

            module = importlib.import_module(module_name)
            script = module.__file__
        return script

    def get_installed_script_path(self, script):
        """
        Returns the path of the executable and the associated environment

        In Windows, it checks availability of script using .py .bat, and .exe
        file extensions.
        """
        if sys.platform == "win32":
            available_extensions = [".py", ".bat", ".exe"]
        else:
            available_extensions = [""]

        paths = os.environ.get("PATH", "").split(os.pathsep)
        for base in paths:
            # clean up extra quotes from paths
            if base.startswith('"') and base.endswith('"'):
                base = base[1:-1]
            for file_extension in available_extensions:
                script_path = os.path.join(base, script + file_extension)
                print(script_path)
                if os.path.exists(script_path):
                    # script found
                    return script_path
        # script not found
        logger.warning("Script '%s' not found in paths: %s", script, ":".join(paths))
        script_path = script
        return script_path

    def _initialize_tmpdir(self):
        """Initialize the temporary directory"""
        if not self._tempdir:
            with self.sem:
                if not self._tempdir:
                    self._tempdir = tempfile.mkdtemp(
                        "_" + getpass.getuser(), self.name + "_"
                    )

    @property
    def tempdir(self):
        if not self._tempdir:
            self._initialize_tmpdir()
        return self._tempdir

    def clean_up(self):
        """Removes the temporary directory (and all its content !)"""
        with self.sem:
            if not self._tempdir:
                return
            if not os.path.isdir(self._tempdir):
                return
            for root, dirs, files in os.walk(self._tempdir, topdown=False):
                for name in files:
                    os.remove(os.path.join(root, name))
                for name in dirs:
                    os.rmdir(os.path.join(root, name))
            os.rmdir(self._tempdir)
            self._tempdir = None


UtilsTest = TestContext()
"""Singleton containing util context of whole the tests"""


class ParametricTestCase(unittest.TestCase):
    pass
